# Xlib.ext.randr -- RandR extension module
#
#    Copyright (C) 2006 Mike Meyer <mwm@mired.org>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,  USA



"""RandR - provide access to the RandR extension information.

This implementation is based off version 1.3 of the XRandR protocol, and may
not be compatible with other versions.

Version 1.2 of the protocol is documented at:
http://cgit.freedesktop.org/xorg/proto/randrproto/tree/randrproto.txt

"""


from Xlib import X
from Xlib.protocol import rq, structs

extname = 'RANDR'


# Event codes #
RRScreenChangeNotify        = 0

# V1.2 additions
RRNotify                    = 1

# RRNotify Subcodes
RRNotify_CrtcChange         = 0
RRNotify_OutputChange       = 1
RRNotify_OutputProperty     = 2


# Event selection bits #
RRScreenChangeNotifyMask    = (1 << 0)

# V1.2 additions
RRCrtcChangeNotifyMask      = (1 << 1)
RROutputChangeNotifyMask    = (1 << 2)
RROutputPropertyNotifyMask  = (1 << 3)


# Constants #
SetConfigSuccess            = 0
SetConfigInvalidConfigTime  = 1
SetConfigInvalidTime        = 2
SetConfigFailed             = 3

# used in the rotation field; rotation and reflection in 0.1 proto.
Rotate_0                    = 1
Rotate_90                   = 2
Rotate_180                  = 4
Rotate_270                  = 8

# new in 1.0 protocol, to allow reflection of screen
Reflect_X                   = 16
Reflect_Y                   = 32

# new in 1.2 protocol
HSyncPositive               = 0x00000001
HSyncNegative               = 0x00000002
VSyncPositive               = 0x00000004
VSyncNegative               = 0x00000008
Interlace                   = 0x00000010
DoubleScan                  = 0x00000020
CSync                       = 0x00000040
CSyncPositive               = 0x00000080
CSyncNegative               = 0x00000100
HSkewPresent                = 0x00000200
BCast                       = 0x00000400
PixelMultiplex              = 0x00000800
DoubleClock                 = 0x00001000
ClockDivideBy2              = 0x00002000

# event types?
Connected                   = 0
Disconnected                = 1
UnknownConnection           = 2

# Conventional RandR output properties
PROPERTY_RANDR_EDID         = "EDID"
PROPERTY_SIGNAL_FORMAT      = "SignalFormat"
PROPERTY_SIGNAL_PROPERTIES  = "SignalProperties"
PROPERTY_CONNECTOR_TYPE     = "ConnectorType"
PROPERTY_CONNECTOR_NUMBER   = "ConnectorNumber"
PROPERTY_COMPATIBILITY_LIST = "CompatibilityList"
PROPERTY_CLONE_LIST         = "CloneList"

# subpixel order - TODO: These constants are part of the RENDER extension and
# should be moved there if/when that extension is added to python-xlib.
SubPixelUnknown             = 0
SubPixelHorizontalRGB       = 1
SubPixelHorizontalBGR       = 2
SubPixelVerticalRGB         = 3
SubPixelVerticalBGR         = 4
SubPixelNone                = 5


# Error Codes #
BadRROutput                 = 0
BadRRCrtc                   = 1
BadRRMode                   = 2


# Data Structures #

RandR_ScreenSizes = rq.Struct(
        rq.Card16('width_in_pixels'),
        rq.Card16('height_in_pixels'),
        rq.Card16('width_in_millimeters'),
        rq.Card16('height_in_millimeters'),
        )


RandR_ModeInfo = rq.Struct(
        rq.Card32('id'),
        rq.Card16('width'),
        rq.Card16('height'),
        rq.Card32('dot_clock'),
        rq.Card16('h_sync_start'),
        rq.Card16('h_sync_end'),
        rq.Card16('h_total'),
        rq.Card16('h_skew'),
        rq.Card16('v_sync_start'),
        rq.Card16('v_sync_end'),
        rq.Card16('v_total'),
        rq.Card16('name_length'),
        rq.Card32('flags'),
        )

RandR_Rates = rq.Struct(
        rq.LengthOf('rates', 2),
        rq.List('rates', rq.Card16Obj)
        )

# TODO: This struct is part of the RENDER extension and should be moved there
# if/when that extension is added to python-xlib.
Render_Transform = rq.Struct(
        rq.Card32('matrix11'), #FIXME: All of these are listed as FIXED in the protocol header.
        rq.Card32('matrix12'),
        rq.Card32('matrix13'),
        rq.Card32('matrix21'),
        rq.Card32('matrix22'),
        rq.Card32('matrix23'),
        rq.Card32('matrix31'),
        rq.Card32('matrix32'),
        rq.Card32('matrix33'),
        )


# Requests #

class QueryVersion(rq.ReplyRequest):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(0),
        rq.RequestLength(),
        rq.Card32('major_version'),
        rq.Card32('minor_version'),
        )
    _reply = rq.Struct(
        rq.ReplyCode(),
        rq.Pad(1),
        rq.Card16('sequence_number'),
        rq.ReplyLength(),
        rq.Card32('major_version'),
        rq.Card32('minor_version'),
        rq.Pad(16),
        )

def query_version(self):
    """Get the current version of the RandR extension.

    """
    return QueryVersion(
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        major_version=1,
        minor_version=3,
        )


class _1_0SetScreenConfig(rq.ReplyRequest):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(2),
        rq.RequestLength(),
        rq.Drawable('drawable'),
        rq.Card32('timestamp'),
        rq.Card32('config_timestamp'),
        rq.Card16('size_id'),
        rq.Card16('rotation'),
        )
    _reply = rq.Struct(
        rq.ReplyCode(),
        rq.Card8('status'),
        rq.Card16('sequence_number'),
        rq.ReplyLength(),
        rq.Card32('new_timestamp'),
        rq.Card32('new_config_timestamp'),
        rq.Window('root'),
        rq.Card16('subpixel_order'),
        rq.Pad(10),
        )

def _1_0set_screen_config(self, size_id, rotation, config_timestamp, timestamp=X.CurrentTime):
    """Sets the screen to the specified size and rotation.

    """
    return _1_0SetScreenConfig(
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        drawable=self,
        timestamp=timestamp,
        config_timestamp=config_timestamp,
        size_id=size_id,
        rotation=rotation,
        )


class SetScreenConfig(rq.ReplyRequest):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(2),
        rq.RequestLength(),
        rq.Drawable('drawable'),
        rq.Card32('timestamp'),
        rq.Card32('config_timestamp'),
        rq.Card16('size_id'),
        rq.Card16('rotation'),
        rq.Card16('rate'), # added in version 1.1
        rq.Pad(2),
        )
    _reply = rq.Struct(
        rq.ReplyCode(),
        rq.Card8('status'),
        rq.Card16('sequence_number'),
        rq.ReplyLength(),
        rq.Card32('new_timestamp'),
        rq.Card32('new_config_timestamp'),
        rq.Window('root'),
        rq.Card16('subpixel_order'),
        rq.Pad(10),
        )

def set_screen_config(self, size_id, rotation, config_timestamp, rate=0, timestamp=X.CurrentTime):
    """Sets the screen to the specified size, rate, rotation and reflection.

    rate can be 0 to have the server select an appropriate rate.

    """
    return SetScreenConfig(
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        drawable=self,
        timestamp=timestamp,
        config_timestamp=config_timestamp,
        size_id=size_id,
        rotation=rotation,
        rate=rate,
        )


class SelectInput(rq.Request):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(4),
        rq.RequestLength(),
        rq.Window('window'),
        rq.Card16('mask'),
        rq.Pad(2),
        )

def select_input(self, mask):
    return SelectInput(
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        window=self,
        mask=mask,
        )


class GetScreenInfo(rq.ReplyRequest):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(5),
        rq.RequestLength(),
        rq.Window('window'),
        )
    _reply = rq.Struct(
        rq.ReplyCode(),
        rq.Card8('set_of_rotations'),
        rq.Card16('sequence_number'),
        rq.ReplyLength(),
        rq.Window('root'),
        rq.Card32('timestamp'),
        rq.Card32('config_timestamp'),
        rq.LengthOf('sizes', 2),
        rq.Card16('size_id'),
        rq.Card16('rotation'),
        rq.Card16('rate'), # added in version 1.1
        rq.Card16('n_rate_ents'), # XCB's protocol description disagrees with the X headers on this; ignoring.
        rq.Pad(2),
        rq.List('sizes', RandR_ScreenSizes),
        #rq.List('rates', RandR_Rates) #FIXME: Why does uncommenting this cause an error?
        )

def get_screen_info(self):
    """Retrieve information about the current and available configurations for
    the screen associated with this window.

    """
    return GetScreenInfo(
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        window=self,
        )


# version 1.2

class GetScreenSizeRange(rq.ReplyRequest):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(6),
        rq.RequestLength(),
        rq.Window('window'),
        )
    _reply = rq.Struct(
        rq.ReplyCode(),
        rq.Pad(1),
        rq.Card16('sequence_number'),
        rq.ReplyLength(),
        rq.Card16('min_width'),
        rq.Card16('min_height'),
        rq.Card16('max_width'),
        rq.Card16('max_height'),
        rq.Pad(16),
        )

def get_screen_size_range(self):
    """Retrieve the range of possible screen sizes. The screen may be set to
	any size within this range.

    """
    return GetScreenSizeRange(
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        window=self,
        )


class SetScreenSize(rq.Request):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(7),
        rq.RequestLength(),
        rq.Window('window'),
        rq.Card16('width'),
        rq.Card16('height'),
        rq.Card32('width_in_millimeters'),
        rq.Card32('height_in_millimeters'),
        )

def set_screen_size(self, width, height, width_in_millimeters=None, height_in_millimeters=None):
    return SetScreenSize(
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        window=self,
        width=width,
        height=height,
        width_in_millimeters=width_in_millimeters,
        height_in_millimeters=height_in_millimeters,
        )


class GetScreenResources(rq.ReplyRequest):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(8),
        rq.RequestLength(),
        rq.Window('window'),
        )
    _reply = rq.Struct(
        rq.ReplyCode(),
        rq.Pad(1),
        rq.Card16('sequence_number'),
        rq.ReplyLength(),
        rq.Card32('timestamp'),
        rq.Card32('config_timestamp'),
        rq.LengthOf('crtcs', 2),
        rq.LengthOf('outputs', 2),
        rq.LengthOf('modes', 2),
        rq.LengthOf('mode_names', 2),
        rq.Pad(8),
        rq.List('crtcs', rq.Card32Obj),
        rq.List('outputs', rq.Card32Obj),
        rq.List('modes', RandR_ModeInfo),
        rq.String8('mode_names'),
        )

def get_screen_resources(self):
    return GetScreenResources(
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        window=self,
        )


class GetOutputInfo(rq.ReplyRequest):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(9),
        rq.RequestLength(),
        rq.Card32('output'),
        rq.Card32('config_timestamp'),
        )
    _reply = rq.Struct(
        rq.ReplyCode(),
        rq.Card8('status'),
        rq.Card16('sequence_number'),
        rq.ReplyLength(),
        rq.Card32('timestamp'),
        rq.Card32('crtc'),
        rq.Card32('mm_width'),
        rq.Card32('mm_height'),
        rq.Card8('connection'),
        rq.Card8('subpixel_order'),
        rq.LengthOf('crtcs', 2),
        rq.LengthOf('modes', 2),
        rq.LengthOf('preferred', 2),
        rq.LengthOf('clones', 2),
        rq.LengthOf('name', 2),
        rq.List('crtcs', rq.Card32Obj),
        rq.List('modes', rq.Card32Obj),
        rq.List('preferred', rq.Card32Obj),
        rq.List('clones', rq.Card32Obj),
        rq.String8('name'),
        )

def get_output_info(self, output, config_timestamp):
    return GetOutputInfo(
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        output=output,
        config_timestamp=config_timestamp,
        )


class ListOutputProperties(rq.ReplyRequest):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(10),
        rq.RequestLength(),
        rq.Card32('output'),
        )
    _reply = rq.Struct(
        rq.ReplyCode(),
        rq.Pad(1),
        rq.Card16('sequence_number'),
        rq.ReplyLength(),
        rq.LengthOf('atoms', 2),
        rq.Pad(22),
        rq.List('atoms', rq.Card32Obj),
        )

def list_output_properties(self, output):
    return ListOutputProperties (
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        output=output,
        )


class QueryOutputProperty(rq.ReplyRequest):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(11),
        rq.RequestLength(),
        rq.Card32('output'),
        rq.Card32('property'),
        )
    _reply = rq.Struct(
        rq.ReplyCode(),
        rq.Pad(1),
        rq.Card16('sequence_number'),
        rq.ReplyLength(),
        rq.Bool('pending'),
        rq.Bool('range'),
        rq.Bool('immutable'),
        rq.Pad(21),
        rq.List('valid_values', rq.Card32Obj),
        )

def query_output_property(self, output, property):
    return QueryOutputProperty (
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        output=output,
        property=property,
        )


class ConfigureOutputProperty (rq.Request):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(12),
        rq.RequestLength(),
        rq.Card32('output'),
        rq.Card32('property'),
        rq.Bool('pending'),
        rq.Bool('range'),
        rq.Pad(2),
        rq.List('valid_values', rq.Card32Obj),
        )

def configure_output_property (self, output, property):
    return ConfigureOutputProperty (
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        output=output,
        property=property,
        )


class ChangeOutputProperty(rq.Request):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(13),
        rq.RequestLength(),
        rq.Card32('output'),
        rq.Card32('property'),
        rq.Card32('type'),
        rq.Format('value', 1),
        rq.Card8('mode'),
        rq.Pad(2),
        rq.LengthOf('value', 4),
        rq.List('value', rq.Card8Obj),
        )

def change_output_property(self, output, property, type, format, mode, nUnits):
    return ChangeOutputProperty(
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        output=output,
        property=property,
        type=type,
        format=format,
        mode=mode,
        nUnits=nUnits,
        )


class DeleteOutputProperty(rq.Request):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(14),
        rq.RequestLength(),
        rq.Card32('output'),
        rq.Card32('property'),
        )

def delete_output_property(self, output, property):
    return DeleteOutputProperty(
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        output=output,
        property=property,
        )


class GetOutputProperty(rq.ReplyRequest):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(15),
        rq.RequestLength(),
        rq.Card32('output'),
        rq.Card32('property'),
        rq.Card32('type'),
        rq.Card32('long_offset'),
        rq.Card32('long_length'),
        rq.Bool('delete'),
        rq.Bool('pending'),
        rq.Pad(2),
        )
    _reply = rq.Struct(
        rq.ReplyCode(),
        rq.Format('value', 1),
        rq.Card16('sequence_number'),
        rq.ReplyLength(),
        rq.Card32('property_type'),
        rq.Card32('bytes_after'),
        rq.LengthOf('value', 4),
        rq.Pad(12),
        rq.List('value', rq.Card8Obj),
        )

def get_output_property(self, output, property, type, longOffset, longLength):
    return GetOutputProperty(
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        output=output,
        property=property,
        type=type,
        longOffset=longOffset,
        longLength=longLength,
        )


class CreateMode(rq.ReplyRequest):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(16),
        rq.RequestLength(),
        rq.Window('window'),
        rq.Object('mode', RandR_ModeInfo),
        rq.String8('name'),
        )
    _reply = rq.Struct(
        rq.ReplyCode(),
        rq.Pad(1),
        rq.Card16('sequence_number'),
        rq.ReplyLength(),
        rq.Card32('mode'),
        rq.Pad(20),
        )

def create_mode(self):
    return CreateMode (
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        window=self,
        )


class DestroyMode(rq.Request):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(17),
        rq.RequestLength(),
        rq.Card32('mode'),
        )

def destroy_mode(self, mode):
    return DestroyMode(
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        mode=mode,
        )


class AddOutputMode(rq.Request):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(18),
        rq.RequestLength(),
        rq.Card32('output'),
        rq.Card32('mode'),
        )

def add_output_mode(self):
    return AddOutputMode(
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        output=output,
        mode=mode,
        )


class DeleteOutputMode(rq.Request):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(19),
        rq.RequestLength(),
        rq.Card32('output'),
        rq.Card32('mode'),
        )

def delete_output_mode(self):
    return DeleteOutputMode(
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        output=output,
        mode=mode,
        )


class GetCrtcInfo(rq.ReplyRequest):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(20),
        rq.RequestLength(),
        rq.Card32('crtc'),
        rq.Card32('config_timestamp'),
        )
    _reply = rq.Struct(
        rq.ReplyCode(),
        rq.Card8('status'),
        rq.Card16('sequence_number'),
        rq.ReplyLength(),
        rq.Card32('timestamp'),
        rq.Card16('width'),
        rq.Card16('height'),
        rq.Card32('mode'),
        rq.Card16('rotation'),
        rq.Card16('possible_rotations'),
        rq.LengthOf('outputs', 2),
        rq.LengthOf('possible_outputs', 2),
        rq.List('outputs', rq.Card32Obj),
        rq.List('possible_outputs', rq.Card32Obj),
        )

def get_crtc_info(self, crtc, config_timestamp):
    return GetCrtcInfo (
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        crtc=crtc,
        config_timestamp=config_timestamp,
        )


class SetCrtcConfig(rq.ReplyRequest):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(21),
        rq.RequestLength(),
        rq.Card32('crtc'),
        rq.Card32('timestamp'),
        rq.Card32('config_timestamp'),
        rq.Int16('x'),
        rq.Int16('y'),
        rq.Card32('mode'),
        rq.Card16('rotation'),
        rq.Pad(2),
        rq.List('outputs', rq.Card32Obj),
        )
    _reply = rq.Struct(
        rq.ReplyCode(),
        rq.Card8('status'),
        rq.Card16('sequence_number'),
        rq.ReplyLength(),
        rq.Card32('new_timestamp'),
        rq.Pad(20),
        )

def set_crtc_config(self, crtc, config_timestamp, mode, rotation, timestamp=X.CurrentTime):
    return SetCrtcConfig (
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        crtc=crtc,
        config_timestamp=config_timestamp,
        mode=mode,
        rotation=rotation,
        timestamp=timestamp,
        )


class GetCrtcGammaSize(rq.ReplyRequest):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(22),
        rq.RequestLength(),
        rq.Card32('crtc'),
        )
    _reply = rq.Struct(
        rq.ReplyCode(),
        rq.Card8('status'),
        rq.Card16('sequence_number'),
        rq.ReplyLength(),
        rq.Card16('size'),
        rq.Pad(22),
        )

def get_crtc_gamma_size(self, crtc):
    return GetCrtcGammaSize (
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        crtc=crtc,
        )


class GetCrtcGamma(rq.ReplyRequest):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(23),
        rq.RequestLength(),
        rq.Card32('crtc'),
        )
    _reply = rq.Struct(
        rq.ReplyCode(),
        rq.Card8('status'),
        rq.Card16('sequence_number'),
        rq.ReplyLength(),
        rq.Card16('size'),
        rq.Pad(22),
        rq.List('red', rq.Card16Obj),
        rq.List('green', rq.Card16Obj),
        rq.List('blue', rq.Card16Obj),
        )

def get_crtc_gamma(self, crtc):
    return GetCrtcGamma (
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        crtc=crtc,
        )


class SetCrtcGamma(rq.Request):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(24),
        rq.RequestLength(),
        rq.Card32('crtc'),
        rq.Card16('size'),
        rq.Pad(2),
        rq.List('red', rq.Card16Obj),
        rq.List('green', rq.Card16Obj),
        rq.List('blue', rq.Card16Obj),
        )

def set_crtc_gamma(self, crtc, size):
    return SetCrtcGamma(
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        crtc=crtc,
        size=size,
        )


# version 1.3

class GetScreenResourcesCurrent(rq.ReplyRequest):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(25),
        rq.RequestLength(),
        rq.Window('window'),
        )
    _reply = rq.Struct(
        rq.ReplyCode(),
        rq.Pad(1),
        rq.Card16('sequence_number'),
        rq.ReplyLength(),
        rq.Card32('timestamp'),
        rq.Card32('config_timestamp'),
        rq.LengthOf('crtcs', 2),
        rq.LengthOf('outputs', 2),
        rq.LengthOf('modes', 2),
        rq.LengthOf('names', 2),
        rq.Pad(8),
        rq.List('crtcs', rq.Card32Obj),
        rq.List('outputs', rq.Card32Obj),
        rq.List('modes', RandR_ModeInfo),
        rq.String8('names'),
        )

def get_screen_resources_current(self):
    return GetScreenResourcesCurrent(
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        window=self,
        )


class SetCrtcTransform(rq.Request):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(26),
        rq.RequestLength(),
        rq.Card32('crtc'),
        rq.Object('transform', Render_Transform),
        rq.LengthOf('filter_name', 2),
        rq.Pad(2),
        rq.String8('filter_name'),
        rq.List('filter_params', rq.Card32Obj), #FIXME: The protocol says FIXED? http://cgit.freedesktop.org/xorg/proto/randrproto/tree/randrproto.txt#n2161
        )

def set_crtc_transform(self, crtc, n_bytes_filter):
    return SetCrtcTransform(
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        crtc=crtc,
        n_bytes_filter=n_bytes_filter,
        )


class GetCrtcTransform(rq.ReplyRequest):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(27),
        rq.RequestLength(),
        rq.Card32('crtc'),
        )
    _reply = rq.Struct(
        rq.ReplyCode(),
        rq.Card8('status'),
        rq.Card16('sequence_number'),
        rq.ReplyLength(),
        rq.Object('pending_transform', Render_Transform),
        rq.Bool('has_transforms'),
        rq.Pad(3),
        rq.Object('current_transform', Render_Transform),
        rq.Pad(4),
        rq.LengthOf('pending_filter_name', 2),
        rq.LengthOf('pending_filter_params', 2),
        rq.LengthOf('current_filter_name', 2),
        rq.LengthOf('current_filter_params', 2),
        rq.String8('pending_filter_name'),
        rq.List('pending_filter_params', rq.Card32Obj), #FIXME: The protocol says FIXED? http://cgit.freedesktop.org/xorg/proto/randrproto/tree/randrproto.txt#n2161
        rq.String8('current_filter_name'),
        rq.List('current_filter_params', rq.Card32Obj), #FIXME: The protocol says FIXED? http://cgit.freedesktop.org/xorg/proto/randrproto/tree/randrproto.txt#n2161
        )

def get_crtc_transform(self, crtc):
    return GetCrtcTransform(
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        crtc=crtc,
        )


class GetPanning(rq.ReplyRequest):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(28),
        rq.RequestLength(),
        rq.Card32('crtc'),
        )
    _reply = rq.Struct(
        rq.ReplyCode(),
        rq.Card8('status'),
        rq.Card16('sequence_number'),
        rq.ReplyLength(),
        rq.Card32('timestamp'),
        rq.Card16('left'),
        rq.Card16('top'),
        rq.Card16('width'),
        rq.Card16('height'),
        rq.Card16('track_left'),
        rq.Card16('track_top'),
        rq.Card16('track_width'),
        rq.Card16('track_height'),
        rq.Int16('border_left'),
        rq.Int16('border_top'),
        rq.Int16('border_right'),
        rq.Int16('border_bottom'),
        )

def get_panning(self, crtc):
    return GetPanning (
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        crtc=crtc,
        )


class SetPanning(rq.ReplyRequest):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(29),
        rq.RequestLength(),
        rq.Card32('crtc'),
        rq.Card32('timestamp'),
        rq.Card16('left'),
        rq.Card16('top'),
        rq.Card16('width'),
        rq.Card16('height'),
        rq.Card16('track_left'),
        rq.Card16('track_top'),
        rq.Card16('track_width'),
        rq.Card16('track_height'),
        rq.Int16('border_left'),
        rq.Int16('border_top'),
        rq.Int16('border_right'),
        rq.Int16('border_bottom'),
        )
    _reply = rq.Struct(
        rq.ReplyCode(),
        rq.Card8('status'),
        rq.Card16('sequence_number'),
        rq.ReplyLength(),
        rq.Card32('new_timestamp'),
        rq.Pad(20),
        )

def set_panning(self, crtc, left, top, width, height, track_left, track_top, track_width, track_height, border_left, border_top, border_width, border_height, timestamp=X.CurrentTime):
    return SetPanning (
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        crtc=crtc,
        left=left,
        top=top,
        width=width,
        height=height,
        track_left=track_left,
        track_top=track_top,
        track_width=track_width,
        track_height=track_height,
        border_left=border_left,
        border_top=border_top,
        border_width=border_width,
        border_height=border_height,
        timestamp=timestamp,
        )


class SetOutputPrimary(rq.Request):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(30),
        rq.RequestLength(),
        rq.Window('window'),
        rq.Card32('output'),
        )

def set_output_primary(self, output):
    return SetOutputPrimary(
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        window=self,
        output=output,
        )


class GetOutputPrimary(rq.ReplyRequest):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(31),
        rq.RequestLength(),
        rq.Window('window'),
        )
    _reply = rq.Struct(
        rq.ReplyCode(),
        rq.Pad(1),
        rq.Card16('sequence_number'),
        rq.ReplyLength(),
        rq.Card32('output'),
        rq.Pad(20),
        )

def get_output_primary(self):
    return GetOutputPrimary(
        display=self.display,
        opcode=self.display.get_extension_major(extname),
        window=self,
        )


# Events #

class ScreenChangeNotify(rq.Event):
    _code = None
    _fields = rq.Struct(
        rq.Card8('type'),
        rq.Card8('rotation'),
        rq.Card16('sequence_number'),
        rq.Card32('timestamp'),
        rq.Card32('config_timestamp'),
        rq.Window('root'),
        rq.Window('window'),
        rq.Card16('size_id'),
        rq.Card16('subpixel_order'),
        rq.Card16('width_in_pixels'),
        rq.Card16('height_in_pixels'),
        rq.Card16('width_in_millimeters'),
        rq.Card16('height_in_millimeters'),
        )


class CrtcChangeNotify(rq.Event):
    _code = None
    _fields = rq.Struct(
        rq.Card8('type'),
        rq.Card8('sub_code'),
        rq.Card16('sequence_number'),
        rq.Card32('timestamp'),
        rq.Window('window'),
        rq.Card32('crtc'),
        rq.Card32('mode'),
        rq.Card16('rotation'),
        rq.Pad(2),
        rq.Int16('x'),
        rq.Int16('y'),
        rq.Card16('width'),
        rq.Card16('height'),
        )


class OutputChangeNotify(rq.Event):
    _code = None
    _fields = rq.Struct(
        rq.Card8('type'),
        rq.Card8('sub_code'),
        rq.Card16('sequence_number'),
        rq.Card32('timestamp'),
        rq.Card32('config_timestamp'),
        rq.Window('window'),
        rq.Card32('output'),
        rq.Card32('crtc'),
        rq.Card32('mode'),
        rq.Card16('rotation'),
        rq.Card8('connection'),
        rq.Card8('subpixel_order'),
        )


class OutputPropertyNotify(rq.Event):
    _code = None
    _fields = rq.Struct(
        rq.Card8('type'),
        rq.Card8('sub_code'),
        rq.Card16('sequence_number'),
        rq.Window('window'),
        rq.Card32('output'),
        rq.Card32('atom'),
        rq.Card32('timestamp'),
        rq.Card8('state'),
        rq.Pad(11),
        )


# Initialization #

def init(disp, info):
    disp.extension_add_method('display', 'xrandr_query_version', query_version)
    disp.extension_add_method('window', 'xrandr_select_input', select_input)
    disp.extension_add_method('window', 'xrandr_get_screen_info', get_screen_info)
    disp.extension_add_method('drawable', 'xrandr_1_0set_screen_config', _1_0set_screen_config)
    disp.extension_add_method('drawable', 'xrandr_set_screen_config', set_screen_config)
    disp.extension_add_method('window', 'xrandr_get_screen_size_range', get_screen_size_range)
    disp.extension_add_method('window', 'xrandr_set_screen_size', set_screen_size)
    disp.extension_add_method('window', 'xrandr_get_screen_resources', get_screen_resources)
    disp.extension_add_method('display', 'xrandr_get_output_info', get_output_info)
    disp.extension_add_method('display', 'xrandr_list_output_properties', list_output_properties)
    disp.extension_add_method('display', 'xrandr_query_output_property', query_output_property)
    disp.extension_add_method('display', 'xrandr_configure_output_property ', configure_output_property )
    disp.extension_add_method('display', 'xrandr_change_output_property', change_output_property)
    disp.extension_add_method('display', 'xrandr_delete_output_property', delete_output_property)
    disp.extension_add_method('display', 'xrandr_get_output_property', get_output_property)
    disp.extension_add_method('window', 'xrandr_create_mode', create_mode)
    disp.extension_add_method('display', 'xrandr_destroy_mode', destroy_mode)
    disp.extension_add_method('display', 'xrandr_add_output_mode', add_output_mode)
    disp.extension_add_method('display', 'xrandr_delete_output_mode', delete_output_mode)
    disp.extension_add_method('display', 'xrandr_get_crtc_info', get_crtc_info)
    disp.extension_add_method('display', 'xrandr_set_crtc_config', set_crtc_config)
    disp.extension_add_method('display', 'xrandr_get_crtc_gamma_size', get_crtc_gamma_size)
    disp.extension_add_method('display', 'xrandr_get_crtc_gamma', get_crtc_gamma)
    disp.extension_add_method('display', 'xrandr_set_crtc_gamma', set_crtc_gamma)
    disp.extension_add_method('window', 'xrandr_get_screen_resources_current', get_screen_resources_current)
    disp.extension_add_method('display', 'xrandr_set_crtc_transform', set_crtc_transform)
    disp.extension_add_method('display', 'xrandr_get_crtc_transform', get_crtc_transform)
    disp.extension_add_method('window', 'xrandr_set_output_primary', set_output_primary)
    disp.extension_add_method('window', 'xrandr_get_output_primary', get_output_primary)
    disp.extension_add_method('display', 'xrandr_get_panning', get_panning)
    disp.extension_add_method('display', 'xrandr_set_panning', set_panning)

    disp.extension_add_event(info.first_event, ScreenChangeNotify)
    disp.extension_add_event(info.first_event + 1, CrtcChangeNotify)
    disp.extension_add_event(info.first_event + 2, OutputChangeNotify)
    disp.extension_add_event(info.first_event + 3, OutputPropertyNotify)

    #disp.extension_add_error(BadRROutput, BadRROutputError)
    #disp.extension_add_error(BadRRCrtc, BadRRCrtcError)
    #disp.extension_add_error(BadRRMode, BadRRModeError)



# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from datetime import datetime

from Nagstamon.cookies import (cookie_data_to_jar,
                               load_cookies,
                               save_cookies)
from Nagstamon.helpers import USER_AGENT
from Nagstamon.qui.qt import (QNetworkProxy,
                              QNetworkCookie,
                              QNetworkProxyFactory,
                              QT_VERSION_MAJOR,
                              QWidget,
                              QUrl,
                              Signal,
                              Slot,
                              WebEngineCertificateError,
                              WebEnginePage,
                              WebEngineView,
                              WebEngineProfile)
from Nagstamon.qui.dialogs.dialog import Dialog
from Nagstamon.servers import servers


class WebEnginePageImproved(WebEnginePage):
    """
    subclass of QWebEnginePage to handle TLS errors and potentially ignore them based on server configuration
    additionaly behave differently in Qt 5 and Qt 6 due to API changes in the certificateError signal/slot mechanism
    """
    def __init__(self, ignore_tls_errors=False, parent=None):
        super().__init__(parent)
        self.ignore_tls_errors = ignore_tls_errors
        if QT_VERSION_MAJOR >= 6:
            self.certificateError.connect(self.handle_certificate_error)
        else:
            self.certificateError = self.handle_certificate_error

    @Slot(WebEngineCertificateError)
    def handle_certificate_error(self, error):
        print('TLS error:', error.description())
        if self.ignore_tls_errors:
            error.acceptCertificate()
            return True
        return False


class DialogWebLogin(Dialog):
    """
    small dialog for web login
    """

    show_up = Signal()
    page_loaded = Signal()
    delete_web_cookies = Signal(str, QWidget)

    def __init__(self):
        Dialog.__init__(self, 'dialog_weblogin', )
        self.cookie_store = None
        self.page = None
        self.profile = None
        self.proxy = None
        self.webengine_view = None
        self.cookies = dict()

    def on_url_changed(self, url):
        """
        change URL label when URL changes
        """
        self.window.label_url.setText(url.toString())

    def on_load_finished(self):
        """
        send message when page is loaded so the statuswindow can refresh
        """
        if self.server:
            self.page_loaded.emit()

    @Slot(str)
    def show_browser(self, server_name):
        """
        initialize and show authentication browser window
        """
        self.show_up.emit()

        self.server = servers[server_name]

        self.proxy = QNetworkProxy()

        if self.server.use_proxy:
            QNetworkProxyFactory.setUseSystemConfiguration(False)
            self.proxy.setType(QNetworkProxy.ProxyType.HttpProxy)
            if self.server.use_proxy_from_os:
                QNetworkProxyFactory.setUseSystemConfiguration(True)
            else:
                self.proxy.setType(QNetworkProxy.ProxyType.HttpProxy)
                # kick out any protocol prefix
                host, port = self.server.proxy_address.split('https://')[-1].split('http://')[-1].split(':')
                # ...and any trailing path
                port = port.split('/')[0]
                self.proxy.setHostName(host)
                self.proxy.setPort(int(port))
                if self.server.proxy_username:
                    self.proxy.setUser(self.server.proxy_username)
                    self.proxy.setPassword(self.server.proxy_password)
                QNetworkProxy.setApplicationProxy(self.proxy)
        else:
            self.proxy.setType(QNetworkProxy.ProxyType.NoProxy)
            QNetworkProxy.setApplicationProxy(self.proxy)

        # reset webengine view if already existing - especially for proxy changes
        if self.page:
            self.page.deleteLater()
            self.page = None
        if self.webengine_view:
            self.webengine_view.close()
            self.window.vbox.removeWidget(self.webengine_view)
            self.webengine_view.destroy()
            self.webengine_view = None
        if self.profile:
            self.profile = None
        if self.cookie_store:
            self.cookie_store = None

        self.webengine_view = WebEngineView()
        self.window.button_reload.clicked.connect(self.webengine_view.reload)
        self.window.button_delete_web_cookies.clicked.connect(self.on_delete_web_cookies)

        self.profile = WebEngineProfile.defaultProfile()
        self.profile.setHttpUserAgent(USER_AGENT)

        self.cookie_store = self.profile.cookieStore()

        # get already stored cookies and add them to the cookie store
        stored_cookies = load_cookies()
        for entry in stored_cookies:
            stored_cookie = stored_cookies[entry]
            # only add cookies which belongs to selected server
            if self.server.name != stored_cookie['server']:
                continue
            cookie = QNetworkCookie()
            # RFCs state that cookie names and values should be ASCII - let's see what happens
            cookie.setName(stored_cookie['name'].encode('ascii', errors='ignore'))
            cookie.setValue(stored_cookie['value'].encode('ascii', errors='ignore'))
            cookie.setDomain(stored_cookie['domain'])
            cookie.setPath(stored_cookie['path'])
            if stored_cookie['expiration']:
                cookie.setExpirationDate(datetime.fromtimestamp(stored_cookie['expiration']))
            cookie.setSecure(stored_cookie['secure'])
            cookie.setHttpOnly(stored_cookie['httponly'])
            self.cookie_store.setCookie(cookie)
        self.cookie_store.cookieAdded.connect(self.handle_cookie_added)

        self.webengine_view.urlChanged.connect(self.on_url_changed)
        self.webengine_view.loadFinished.connect(self.on_load_finished)

        self.window.vbox.addWidget(self.webengine_view)

        self.window.setWindowTitle('Nagstamon Web Login - ' + server_name)
        self.page = WebEnginePageImproved(ignore_tls_errors=self.server.ignore_cert)
        self.webengine_view.setPage(self.page)
        self.page.setUrl(QUrl(self.server.monitor_url))

        self.show()
        self.window.adjustSize()

        # the dock icon might be needed to be shown for a potential keyboard input
        self.check_macos_dock_icon_fix_show.emit()

        self.window.open()

        # en reverse the dock icon might be hidden again after a potential keyboard input
        self.check_macos_dock_icon_fix_hide.emit()

        self.webengine_view.show()
        self.window.update()

    @Slot(str)
    def close_browser(self, server_name):
        """
        close browser window
        """
        # according to https://bugreports.qt.io/browse/QTBUG-128345 the QWebEngine can't be closed, so it will
        # run further in the background
        # this might be mitigated by only needing it once for initial login and cookie retrieval
        if hasattr(self, 'server') and \
                self.server.name == server_name:
            self.window.close()
            if self.page:
                self.page.deleteLater()
                self.page = None
            if self.webengine_view:
                self.window.vbox.removeWidget(self.webengine_view)
                self.webengine_view.destroy()
                self.webengine_view = None
            if self.profile:
                self.profile = None
            if self.cookie_store:
                self.cookie_store = None

    def handle_cookie_added(self, cookie):
        # load existing cookies from SQLite databases
        cookies = load_cookies()
        # extract relevant cookie data as dictionary
        cookie_data = {
            'name': cookie.name().data().decode(),
            'value': cookie.value().data().decode(),
            'domain': cookie.domain(),
            'path': cookie.path(),
            'expiration': cookie.expirationDate().toSecsSinceEpoch() if cookie.expirationDate().isValid() else None,
            'secure': cookie.isSecure(),
            'httponly': cookie.isHttpOnly(),
            'server': self.server.name
        }
        # Save cookie only if not already saved
        cookey = f"{self.server.name}+{cookie_data['domain']}+{cookie_data['path']}+{cookie_data['name']}"
        if cookie_data not in cookies.values():
            cookies[cookey] = cookie_data
            save_cookies(cookies)

        # update server cookies to jar format
        self.server.session.cookies = cookie_data_to_jar(self.server.name, cookies)

    def on_delete_web_cookies(self):
        """
        delete all cookies for given server
        """
        self.delete_web_cookies.emit(self.server.name, self)


# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from os import (environ,
                sep)
from sys import stdout
from traceback import print_exc

from Nagstamon.config import (conf,
                              debug_queue,
                              OS,
                              OS_MACOS,
                              RESOURCES, OS_NON_LINUX)
from Nagstamon.qui.globals import (resource_files,
                                   statuswindow_properties)
from Nagstamon.qui.qt import (QCursor,
                              QMenu,
                              QPainter,
                              QPixmap,
                              QSvgRenderer,
                              QSystemTrayIcon,
                              Qt,
                              QTimer,
                              QXmlStreamReader,
                              Signal,
                              Slot)
from Nagstamon.qui.widgets.icon import QIconWithFilename
from Nagstamon.servers import get_worst_status


class SystemTrayIcon(QSystemTrayIcon):
    """
    Icon in system tray, works at least in Windows and OSX
    Several Linux desktop environments have different problems

    For some dark, very dark reason systray menu does NOT work in
    Windows if run on commandline as nagstamon.py - the binary .exe works
    """
    show_popwin = Signal()
    hide_popwin = Signal()

    # flag for displaying error icon in case of error
    error_shown = False

    def __init__(self):
        # debug environment variables
        if conf.debug_mode:
            for environment_key, environment_value in environ.items():
                debug_queue.append(f'DEBUG: Environment variable: {environment_key}={environment_value}')

        # initialize systray icon
        QSystemTrayIcon.__init__(self)

        # icons are in dictionary
        self.icons = {}
        self.create_icons()
        # empty icon for flashing notification
        self.icons['EMPTY'] = QIconWithFilename(f'{RESOURCES}{sep}nagstamon_systrayicon_empty.svg')
        # little workaround to match statuswindow.worker_notification.worst_notification_status
        self.icons['UP'] = self.icons['OK']
        # default icon is OK
        if conf.icon_in_systray:
            self.setIcon(self.icons['OK'])

        # store icon for flashing
        self.current_icon = None

        # no menu at first
        self.menu = None

        # timer for singleshots for flashing
        self.timer = QTimer()

        # treat clicks
        self.activated.connect(self.icon_clicked)

    def current_icon_name(self):
        """
        internal function useful for debugging, returns the name of the
        current icon
        """
        current_account_icon = self.icon()
        if current_account_icon is None:
            return '<none>'
        return str(current_account_icon)

    @Slot(QMenu)
    def set_menu(self, menu):
        """
        create current menu for right clicks
        """
        # store menu for future use, especially for MacOSX
        self.menu = menu

        # MacOSX does not distinguish between left and right click so menu will go to upper menu bar
        # update: apparently not, but own context menu will be shown when icon is clicked an all is OK = green
        if OS != OS_MACOS:
            self.setContextMenu(self.menu)

    @Slot()
    def create_icons(self):
        """
        create icons from template, applying colors
        """
        svg_template = f'{RESOURCES}{sep}nagstamon_systrayicon_template.svg'
        # get template from file
        # by using RESOURCE_FILES the file path will be checked on macOS and the file restored if necessary
        with open(resource_files[svg_template]) as svg_template_file:
            svg_template_xml = svg_template_file.readlines()

            # create icons for all states
            for state in ['OK', 'INFORMATION', 'UNKNOWN', 'WARNING', 'AVERAGE', 'HIGH', 'CRITICAL', 'DISASTER',
                          'UNREACHABLE', 'DOWN', 'ERROR']:
                # current SVG XML for state icon, derived from svg_template_cml
                svg_state_xml = list()

                # replace dummy text and background colors with configured ones
                for line in svg_template_xml:
                    line = line.replace('fill:#ff00ff', 'fill:' + conf.__dict__['color_' + state.lower() + '_text'])
                    line = line.replace('fill:#00ff00',
                                        'fill:' + conf.__dict__['color_' + state.lower() + '_background'])
                    svg_state_xml.append(line)

                # create XML stream of SVG
                svg_xml_stream = QXmlStreamReader(''.join(svg_state_xml))
                # create renderer for SVG and put SVG XML into renderer
                svg_renderer = QSvgRenderer(svg_xml_stream)
                # pixmap to be painted on - arbitrarily choosen 128x128 px
                svg_pixmap = QPixmap(128, 128)
                # fill transparent backgound
                svg_pixmap.fill(Qt.GlobalColor.transparent)
                # initiate painter which paints onto paintdevice pixmap
                svg_painter = QPainter(svg_pixmap)
                # render svg to pixmap
                svg_renderer.render(svg_painter)
                # close painting
                svg_painter.end()
                # put pixmap into icon
                self.icons[state] = QIconWithFilename(svg_pixmap)

                debug_queue.append(f'DEBUG: SystemTrayIcon created icon {self.icons[state]} for state "{state}"')

    @Slot(QSystemTrayIcon.ActivationReason)
    def icon_clicked(self, reason):
        """
        evaluate mouse click
        """
        # retrieve icon position and store it in statuswindow_properties
        self.retrieve_icon_position()

        if reason in (QSystemTrayIcon.ActivationReason.Trigger,
                      QSystemTrayIcon.ActivationReason.DoubleClick,
                      QSystemTrayIcon.ActivationReason.MiddleClick):
            # when green icon is displayed and no popwin is about to pop up...
            if get_worst_status() == 'UP':
                # ...nothing to do except on macOS where menu should be shown
                if OS == OS_MACOS:
                    # in case there is some error show popwin rather than context menu
                    if not self.error_shown:
                        self.menu.show_at_cursor()
                    else:
                        self.show_popwin.emit()
            else:
                # show status window if there is something to tell
                if statuswindow_properties.is_shown:
                    self.hide_popwin.emit()
                else:
                    self.show_popwin.emit()

    @Slot()
    def retrieve_icon_position(self):
        """
        get the coordinates of the systray icon and store it in statuswindow_properties
        """
        # where is the pointer which clicked onto systray icon
        icon_x = self.geometry().x()
        icon_y = self.geometry().y()
        if OS in OS_NON_LINUX:
            if statuswindow_properties.icon_x == 0:
                statuswindow_properties.icon_x = QCursor.pos().x()
            elif icon_x != 0:
                statuswindow_properties.icon_x = icon_x
        else:
            # strangely enough on KDE the systray icon geometry gives back 0, 0 as coordinates
            # also at Ubuntu Unity 16.04
            if icon_x == 0 and statuswindow_properties.icon_x == 0:
                statuswindow_properties.icon_x = QCursor.pos().x()
            elif icon_x != 0:
                statuswindow_properties.icon_x = icon_x

        if icon_y == 0 and statuswindow_properties.icon_y == 0:
            statuswindow_properties.icon_y = QCursor.pos().y()

        if OS in OS_NON_LINUX:
            if statuswindow_properties.icon_y == 0:
                statuswindow_properties.icon_y = QCursor.pos().y()
            elif icon_y != 0:
                statuswindow_properties.icon_y = icon_y

        pass

    @Slot()
    def show_state(self):
        """
        get the worst status and display it in systray
        """
        if not self.error_shown:
            worst_status = get_worst_status()
            self.setIcon(self.icons[worst_status])
            # set current icon for flashing
            self.current_icon = self.icons[worst_status]
            del worst_status
        else:
            self.setIcon(self.icons['ERROR'])

    @Slot()
    def flash(self):
        """
        send color inversion signal to labels
        """
        # only if currently a notification is necessary
        if statuswindow_properties.is_notifying:
            # store current icon to get it reset back
            if self.current_icon is None:
                if not self.error_shown:
                    self.current_icon = self.icons[statuswindow_properties.worst_notification_status]
                else:
                    self.current_icon = self.icons['ERROR']
            # use empty SVG icon to display emptiness
            if resource_files[self.icons['EMPTY'].filename]:
                self.setIcon(self.icons['EMPTY'])
            # fire up a singleshot to reset color soon
            self.timer.singleShot(500, self.reset)

    @Slot()
    def reset(self):
        """
        tell labels to set original colors
        """
        # only if currently a notification is necessary
        if statuswindow_properties.is_notifying:
            try:
                # set curent status icon
                self.setIcon(self.current_icon)
                # even later call itself to invert colors as flash
                self.timer.singleShot(500, self.flash)
            except:
                print_exc(file=stdout)
        else:
            if self.current_icon is not None:
                self.setIcon(self.current_icon)
            self.current_icon = None

    @Slot()
    def set_error(self):
        self.error_shown = True

    @Slot()
    def reset_error(self):
        self.error_shown = False



"""This module intends to provide an implementation of Extended Window Manager Hints, based on
the Xlib modules for python.

See the freedesktop.org `specification <http://standards.freedesktop.org/wm-spec/wm-spec-latest.html>`_ for more information.
"""

from Xlib import display, X, protocol, Xatom
import time

class EWMH:
	"""This class provides the ability to get and set properties defined by the EWMH spec.
	
	Each property can be accessed in two ways. For example, to get the active window::

	  win = ewmh.getActiveWindow()
	  # or: win = ewmh.getProperty('_NET_ACTIVE_WINDOW')
	
	Similarly, to set the active window::

	  ewmh.setActiveWindow(myWindow)
	  # or: ewmh.setProperty('_NET_ACTIVE_WINDOW', myWindow)
	
	When a property is written, don't forget to really send the notification by flushing requests::

	  ewmh.display.flush()
	
	:param _display: the display to use. If not given, Xlib.display.Display() is used.
	:param root: the root window to use. If not given, self.display.screen().root is used."""
	
	NET_WM_WINDOW_TYPES = (
		'_NET_WM_WINDOW_TYPE_DESKTOP', '_NET_WM_WINDOW_TYPE_DOCK', '_NET_WM_WINDOW_TYPE_TOOLBAR', '_NET_WM_WINDOW_TYPE_MENU',
		'_NET_WM_WINDOW_TYPE_UTILITY', '_NET_WM_WINDOW_TYPE_SPLASH', '_NET_WM_WINDOW_TYPE_DIALOG', '_NET_WM_WINDOW_TYPE_DROPDOWN_MENU',
		'_NET_WM_WINDOW_TYPE_POPUP_MENU', '_NET_WM_WINDOW_TYPE_NOTIFICATION', '_NET_WM_WINDOW_TYPE_COMBO', '_NET_WM_WINDOW_TYPE_DND',
		'_NET_WM_WINDOW_TYPE_NORMAL')
	"""List of strings representing all known window types."""
	
	NET_WM_ACTIONS = (
		'_NET_WM_ACTION_MOVE', '_NET_WM_ACTION_RESIZE', '_NET_WM_ACTION_MINIMIZE', '_NET_WM_ACTION_SHADE',
		'_NET_WM_ACTION_STICK', '_NET_WM_ACTION_MAXIMIZE_HORZ', '_NET_WM_ACTION_MAXIMIZE_VERT', '_NET_WM_ACTION_FULLSCREEN',
		'_NET_WM_ACTION_CHANGE_DESKTOP', '_NET_WM_ACTION_CLOSE', '_NET_WM_ACTION_ABOVE', '_NET_WM_ACTION_BELOW')
	"""List of strings representing all known window actions."""
	
	NET_WM_STATES = (
		'_NET_WM_STATE_MODAL', '_NET_WM_STATE_STICKY', '_NET_WM_STATE_MAXIMIZED_VERT', '_NET_WM_STATE_MAXIMIZED_HORZ',
		'_NET_WM_STATE_SHADED', '_NET_WM_STATE_SKIP_TASKBAR', '_NET_WM_STATE_SKIP_PAGER', '_NET_WM_STATE_HIDDEN',
		'_NET_WM_STATE_FULLSCREEN','_NET_WM_STATE_ABOVE', '_NET_WM_STATE_BELOW', '_NET_WM_STATE_DEMANDS_ATTENTION')
	"""List of strings representing all known window states."""
	
	def __init__(self, _display=None, root = None):
		self.display = _display or display.Display()
		self.root = root or self.display.screen().root
		self.__getAttrs = {
			'_NET_CLIENT_LIST': 		self.getClientList,
			'_NET_CLIENT_LIST_STACKING':	self.getClientListStacking,
			'_NET_NUMBER_OF_DESKTOPS':	self.getNumberOfDesktops,
			'_NET_DESKTOP_GEOMETRY':	self.getDesktopGeometry,
			'_NET_DESKTOP_VIEWPORT':	self.getDesktopViewPort,
			'_NET_CURRENT_DESKTOP':		self.getCurrentDesktop,
			'_NET_ACTIVE_WINDOW':		self.getActiveWindow,
			'_NET_WORKAREA':		self.getWorkArea,
			'_NET_SHOWING_DESKTOP':		self.getShowingDesktop,
			'_NET_WM_NAME':			self.getWmName,
			'_NET_WM_VISIBLE_NAME':		self.getWmVisibleName,
			'_NET_WM_DESKTOP':		self.getWmDesktop,
			'_NET_WM_WINDOW_TYPE':		self.getWmWindowType,
			'_NET_WM_STATE':		self.getWmState, 
			'_NET_WM_ALLOWED_ACTIONS':	self.getWmAllowedActions,
			'_NET_WM_PID':			self.getWmPid,
		}
		self.__setAttrs = {
			'_NET_NUMBER_OF_DESKTOPS':	self.setNumberOfDesktops,
			'_NET_DESKTOP_GEOMETRY':	self.setDesktopGeometry,
			'_NET_DESKTOP_VIEWPORT':	self.setDesktopViewport,
			'_NET_CURRENT_DESKTOP':		self.setCurrentDesktop,
			'_NET_ACTIVE_WINDOW':		self.setActiveWindow,
			'_NET_SHOWING_DESKTOP':		self.setShowingDesktop,
			'_NET_CLOSE_WINDOW':		self.setCloseWindow,
			'_NET_MOVERESIZE_WINDOW':	self.setMoveResizeWindow,
			'_NET_WM_NAME':			self.setWmName,
			'_NET_WM_VISIBLE_NAME':		self.setWmVisibleName,
			'_NET_WM_DESKTOP':		self.setWmDesktop,
			'_NET_WM_STATE':		self.setWmState, 
		}
	
	# ------------------------ setters properties ------------------------
	
	def setNumberOfDesktops(self, nb):
		"""Set the number of desktops (property _NET_NUMBER_OF_DESKTOPS).
		
		:param nb: the number of desired desktops"""
		self._setProperty('_NET_NUMBER_OF_DESKTOPS', [nb])
	
	def setDesktopGeometry(self, w, h):
		"""Set the desktop geometry (property _NET_DESKTOP_GEOMETRY)
		
		:param w: desktop width
		:param h: desktop height"""
		self._setProperty('_NET_DESKTOP_GEOMETRY', [w, h])
	
	def setDesktopViewport(self, w, h):
		"""Set the viewport size of the current desktop (property _NET_DESKTOP_VIEWPORT)
		
		:param w: desktop width
		:param h: desktop height"""
		self._setProperty('_NET_DESKTOP_VIEWPORT', [w, h])
	
	def setCurrentDesktop(self, i):
		"""Set the current desktop (property _NET_CURRENT_DESKTOP).
		
		:param i: the desired desktop number"""
		self._setProperty('_NET_CURRENT_DESKTOP', [i, X.CurrentTime])
	
	def setActiveWindow(self, win):
		"""Set the given window active (property _NET_ACTIVE_WINDOW)
		
		:param win: the window object"""
		self._setProperty('_NET_ACTIVE_WINDOW', [1, X.CurrentTime, win.id], win)
	
	def setShowingDesktop(self, show):
		"""Set/unset the mode Showing desktop (property _NET_SHOWING_DESKTOP)
		
		:param show: 1 to set the desktop mode, else 0"""
		self._setProperty('_NET_SHOWING_DESKTOP', [show])
	
	def setCloseWindow(self, win):
		"""Colse the given window (property _NET_CLOSE_WINDOW)
		
		:param win: the window object"""
		self._setProperty('_NET_CLOSE_WINDOW', [int(time.mktime(time.localtime())), 1], win)
	
	def setWmName(self, win, name):
		"""Set the property _NET_WM_NAME
		
		:param win: the window object
		:param name: desired name"""
		self._setProperty('_NET_WM_NAME', name, win)
	
	def setWmVisibleName(self, win, name):
		"""Set the property _NET_WM_VISIBLE_NAME
		
		:param win: the window object
		:param name: desired visible name"""
		self._setProperty('_NET_WM_VISIBLE_NAME', name, win)
	
	def setWmDesktop(self, win, i):
		"""Move the window to the desired desktop by changing the property _NET_WM_DESKTOP.
		
		:param win: the window object
		:param i: desired desktop number"""
		self._setProperty('_NET_WM_DESKTOP', [i, 1], win)
	
	def setMoveResizeWindow(self, win, gravity=0, x=None, y=None, w=None, h=None):
		"""Set the property _NET_MOVERESIZE_WINDOW to move or resize the given window.
		Flags are automatically calculated if x, y, w or h are defined.
		
		:param win: the window object
		:param gravity: gravity (one of the Xlib.X.*Gravity constant or 0)
		:param x: int or None
		:param y: int or None
		:param w: int or None
		:param h: int or None"""
		gravity_flags = gravity | 0b0000100000000000 # indicate source (application)
		if x is None: x = 0
		else: gravity_flags = gravity_flags | 0b0000010000000000 # indicate presence of x
		if y is None: y = 0
		else: gravity_flags = gravity_flags | 0b0000001000000000 # indicate presence of y
		if w is None: w = 0
		else: gravity_flags = gravity_flags | 0b0000000100000000 # indicate presence of w
		if h is None: h = 0
		else: gravity_flags = gravity_flags | 0b0000000010000000 # indicate presence of h
		self._setProperty('_NET_MOVERESIZE_WINDOW', [gravity_flags, x, y, w, h], win)
	
	def setWmState(self, win, action, state, state2=0):
		"""Set/unset one or two state(s) for the given window (property _NET_WM_STATE).
		
		:param win: the window object
		:param action: 0 to remove, 1 to add or 2 to toggle state(s)
		:param state: a state
		:type state: int or str (see :attr:`NET_WM_STATES`)
		:param state2: a state or 0
		:type state2: int or str (see :attr:`NET_WM_STATES`)"""
		if type(state) != int: state = self.display.get_atom(state, 1)
		if type(state2) != int: state2 = self.display.get_atom(state2, 1)
		self._setProperty('_NET_WM_STATE', [action, state, state2, 1], win)
	
	# ------------------------ getters properties ------------------------
	
	def getClientList(self):
		"""Get the list of windows maintained by the window manager for the property
		_NET_CLIENT_LIST.
		
		:return: list of Window objects"""
		return map(self._createWindow, self._getProperty('_NET_CLIENT_LIST'))
	
	def getClientListStacking(self):
		"""Get the list of windows maintained by the window manager for the property
		_NET_CLIENT_LIST_STACKING.
		
		:return: list of Window objects"""
		return map(self._createWindow, self._getProperty('_NET_CLIENT_LIST_STACKING'))
	
	def getNumberOfDesktops(self):
		"""Get the number of desktops (property _NET_NUMBER_OF_DESKTOPS).
		
		:return: int"""
		return self._getProperty('_NET_NUMBER_OF_DESKTOPS')[0]
	
	def getDesktopGeometry(self):
		"""Get the desktop geometry (property _NET_DESKTOP_GEOMETRY) as an array
		of two integers [width, height].
		
		:return: [int, int]"""
		return self._getProperty('_NET_DESKTOP_GEOMETRY')
	
	def getDesktopViewPort(self):
		"""Get the current viewports of each desktop as a list of [x, y] representing
                the top left corner (property _NET_DESKTOP_VIEWPORT).
		
		:return: list of [int, int]"""
		return self._getProperty('_NET_DESKTOP_VIEWPORT')
	
	def getCurrentDesktop(self):
		"""Get the current desktop number (property _NET_CURRENT_DESKTOP)
		
		:return: int"""
		return self._getProperty('_NET_CURRENT_DESKTOP')[0]
	
	def getActiveWindow(self):
		"""Get the current active (toplevel) window or None (property _NET_ACTIVE_WINDOW)
		
		:return: Window object or None"""
		active_window = self._getProperty('_NET_ACTIVE_WINDOW')
		if active_window == None:
			return None
		return self._createWindow(active_window[0])
	
	def getWorkArea(self):
		"""Get the work area for each desktop (property _NET_WORKAREA) as a list of [x, y, width, height]
		
		:return: a list of [int, int, int, int]"""
		return self._getProperty('_NET_WORKAREA')
	
	def getShowingDesktop(self):
		"""Get the value of "showing the desktop" mode of the window manager (property _NET_SHOWING_DESKTOP).
		1 means the mode is activated, and 0 means deactivated.
		
		:return: int"""
		return self._getProperty('_NET_SHOWING_DESKTOP')[0]
	
	def getWmName(self, win):
		"""Get the property _NET_WM_NAME for the given window as a string.
		
		:param win: the window object
		:return: str"""
		return self._getProperty('_NET_WM_NAME', win)
	
	def getWmVisibleName(self, win):
		"""Get the property _NET_WM_VISIBLE_NAME for the given window as a string.
		
		:param win: the window object
		:return: str"""
		return self._getProperty('_NET_WM_VISIBLE_NAME', win)
	
	def getWmDesktop(self, win):
		"""Get the current desktop number of the given window (property _NET_WM_DESKTOP).
		
		:param win: the window object
		:return: int"""
		return self._getProperty('_NET_WM_DESKTOP', win)[0]
	
	def getWmWindowType(self, win, str=False):
		"""Get the list of window types of the given window (property _NET_WM_WINDOW_TYPE).
		
		:param win: the window object
		:param str: True to get a list of string types instead of int
		:return: list of (int|str)"""
		types = self._getProperty('_NET_WM_WINDOW_TYPE', win)
		if not str: return types
		return map(self._getAtomName, wtypes)
	
	def getWmState(self, win, str=False):
		"""Get the list of states of the given window (property _NET_WM_STATE).
		
		:param win: the window object
		:param str: True to get a list of string states instead of int
		:return: list of (int|str)"""
		states = self._getProperty('_NET_WM_STATE', win)
		if not str: return states
		return map(self._getAtomName, states)
	
	def getWmAllowedActions(self, win, str=False):
		"""Get the list of allowed actions for the given window (property _NET_WM_ALLOWED_ACTIONS).
		
		:param win: the window object
		:param str: True to get a list of string allowed actions instead of int
		:return: list of (int|str)"""
		wAllowedActions = self._getProperty('_NET_WM_ALLOWED_ACTIONS', win)
		if not str: return wAllowedActions	
		return map(self._getAtomName, wAllowedActions)
	
	def getWmPid(self, win):
		"""Get the pid of the application associated to the given window (property _NET_WM_PID)
		
		:param win: the window object"""
		return self._getProperty('_NET_WM_PID', win)[0]
	
	def _getProperty(self, _type, win=None):
		if not win: win = self.root
		atom = win.get_full_property(self.display.get_atom(_type), X.AnyPropertyType)
		if atom: return atom.value
	
	def _setProperty(self, _type, data, win=None, mask=None):
		"""Send a ClientMessage event to the root window"""
		if not win: win = self.root
		if type(data) is str:
			dataSize = 8
		else:
			data = (data+[0]*(5-len(data)))[:5]
			dataSize = 32
		
		ev = protocol.event.ClientMessage(window=win, client_type=self.display.get_atom(_type), data=(dataSize, data))

		if not mask:
			mask = (X.SubstructureRedirectMask|X.SubstructureNotifyMask)
		self.root.send_event(ev, event_mask=mask)
	
	def _getAtomName(self, atom):
		try: return self.display.get_atom_name(atom)
		except: return 'UNKNOWN'
	
	def _createWindow(self, wId):
		if not wId: return None
		return self.display.create_resource_object('window', wId)

	def getReadableProperties(self):
		"""Get all the readable properties' names"""
		return self.__getAttrs.keys()

	def getProperty(self, prop, *args, **kwargs):
		"""Get the value of a property. See the corresponding method for the required arguments.
		For example, for the property _NET_WM_STATE, look for :meth:`getWmState`"""
		f = self.__getAttrs.get(prop)
		if not f: raise KeyError('Unknow readable property: %s' % prop)
		return f(self, *args, **kwargs)
	
	def getWritableProperties(self):
		"""Get all the writable properties names"""
		return self.__setAttrs.keys()
	
	def setProperty(self, prop, *args, **kwargs):
		"""Set the value of a property by sending an event on the root window. See the corresponding method for
		the required arguments. For example, for the property _NET_WM_STATE, look for :meth:`setWmState`"""
		f = self.__setAttrs.get(prop)
		if not f: raise KeyError('Unknow writable property: %s' % prop)
		f(self, *args, **kwargs)



# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from Nagstamon.config import conf
from Nagstamon.qui.dialogs.dialog import Dialog
from Nagstamon.qui.qt import (Qt,
                              Signal,
                              Slot)
from Nagstamon.servers import servers

class DialogAuthentication(Dialog):
    """
    dialog for authentication
    """
    # signal for telling server_vbox label to update
    update = Signal(str)

    # signal to tell the world that the authentication dialog will show up
    show_up = Signal()

    def __init__(self):
        Dialog.__init__(self, 'dialog_authentication')
        # store server
        self.server = None

    def initialize(self):
        """
        setup dialog fitting to server
        """
        if self.server is not None:

            self.window.setWindowTitle(f'Authenticate {self.server.name}')
            if self.server.type in ['Centreon', 'Thruk']:
                self.window.input_checkbox_use_autologin.show()
                self.window.input_lineedit_autologin_key.show()
                self.window.input_lineedit_autologin_key.show()
                self.window.label_autologin_key.show()
                # enable switching autologin key and password
                self.window.input_checkbox_use_autologin.clicked.connect(self.toggle_autologin)
                self.window.input_checkbox_use_autologin.setChecked(self.server.use_autologin)
                self.window.input_lineedit_autologin_key.setText(self.server.autologin_key)
                # initialize autologin
                self.toggle_autologin()
            else:
                self.window.input_checkbox_use_autologin.hide()
                self.window.input_lineedit_autologin_key.hide()
                self.window.label_autologin_key.hide()

            # set existing values
            self.window.input_lineedit_username.setText(self.server.username)
            self.window.input_lineedit_password.setText(self.server.password)
            self.window.input_checkbox_save_password.setChecked(conf.servers[self.server.name].save_password)

    @Slot(str)
    def show_auth_dialog(self, server):
        """
        initialize and show authentication dialog
        """
        self.server = servers[server]
        self.initialize()
        self.show_up.emit()
        self.window.adjustSize()

        # the dock icon might be needed to be shown for a potential keyboard input
        self.check_macos_dock_icon_fix_show.emit()

        self.window.exec()

        # en reverse the dock icon might be hidden again after a potential keyboard input
        self.check_macos_dock_icon_fix_hide.emit()

    def ok(self):
        """
        take username and password
        """

        # close window fist to avoid lagging UI
        self.window.close()

        self.server.username = self.window.input_lineedit_username.text()
        self.server.password = self.window.input_lineedit_password.text()
        self.server.refresh_authentication = False

        # store password if it should be saved
        if self.window.input_checkbox_save_password.isChecked():
            conf.servers[self.server.name].username = self.server.username
            conf.servers[self.server.name].password = self.server.password
            conf.servers[self.server.name].save_password = self.window.input_checkbox_save_password.isChecked()
            # store server settings
            conf.save_multiple_config('servers', 'server')

        # Centreon
        if self.server.type in ['Centreon', 'Thruk']:
            if self.window.input_checkbox_use_autologin:
                conf.servers[self.server.name].use_autologin = self.window.input_checkbox_use_autologin.isChecked()
                conf.servers[self.server.name].autologin_key = self.window.input_lineedit_autologin_key.text()
                # store server settings
                conf.save_multiple_config('servers', 'server')

        # reset server connection
        self.server.reset_http()

        # force server to recheck right now
        self.server.thread_counter = conf.update_interval_seconds

        # update server_vbox label
        self.update.emit(self.server.name)

        # call close and macOS dock icon treatment from ancestor
        super().ok()

    @Slot()
    def toggle_autologin(self):
        """
        toggle autologin option for Centreon
        """
        if self.window.input_checkbox_use_autologin.isChecked():
            self.window.label_username.hide()
            self.window.label_password.hide()
            self.window.input_lineedit_username.hide()
            self.window.input_lineedit_password.hide()
            self.window.input_checkbox_save_password.hide()

            self.window.label_autologin_key.show()
            self.window.input_lineedit_autologin_key.show()
        else:
            self.window.label_username.show()
            self.window.label_password.show()
            self.window.input_lineedit_username.show()
            self.window.input_lineedit_password.show()
            self.window.input_checkbox_save_password.show()

            self.window.label_autologin_key.hide()
            self.window.input_lineedit_autologin_key.hide()

        # adjust dialog window size after UI changes
        self.window.adjustSize()



# encoding: utf-8

# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

import arrow
import copy
import datetime
import json
import logging
import sys
import dateutil.parser
import urllib.parse

from Nagstamon.config import conf
from Nagstamon.servers.Generic import GenericServer
from Nagstamon.objects import (GenericHost, GenericService, Result)

log = logging.getLogger(__name__)


class Icinga2APIServer(GenericServer):
    """
        object of Icinga2 server API
    """
    TYPE = 'Icinga2API'

    # ICINGA2API does not provide a web interface for humans
    MENU_ACTIONS = ['Recheck', 'Acknowledge', 'Submit check result', 'Downtime']
    STATES_MAPPING = {'hosts' : {0 : 'UP', 1 : 'DOWN', 2 : 'UNREACHABLE'}, \
                     'services' : {0 : 'OK', 1 : 'WARNING', 2 : 'CRITICAL', 3 : 'UNKNOWN'}}
    STATES_MAPPING_REV = {'hosts' : { 'UP': 0, 'DOWN': 1, 'UNREACHABLE': 2}, \
                         'services' : {'OK': 0, 'WARNING': 1, 'CRITICAL': 2, 'UNKNOWN': 3}}
    BROWSER_URLS = {}


    def __init__(self, **kwds):
        """
        Prepare all urls needed by nagstamon and icinga
        """
        GenericServer.__init__(self, **kwds)

        self.url = conf.servers[self.get_name()].monitor_url
        self.username = conf.servers[self.get_name()].username
        self.password = conf.servers[self.get_name()].password

    def _insert_service_to_hosts(self, service: GenericService):
        """
        We want to create hosts for faulty services as GenericService requires
        that logic.
        """
        service_host = service.get_host_name()
        if service_host not in self.new_hosts:
            self.new_hosts[service_host] = GenericHost()
            self.new_hosts[service_host].name = service_host
            self.new_hosts[service_host].site = service.site
        self.new_hosts[service_host].services[service.name] = service

    def _get_status(self):
        """
            Get status from Icinga Server and translate it into Nagstamon magic
            generic array
        """
        # new_hosts dictionary
        self.new_hosts = dict()

        # hosts - the down ones
        try:
            # We ask icinga for hosts which are not doing well
            hosts = self._get_host_events()
            assert isinstance(hosts, list), "Fail to list hosts"
            for host in hosts:
                host_name = host['attrs']['name']
                if host_name not in self.new_hosts:
                    self.new_hosts[host_name] = GenericHost()
                    self.new_hosts[host_name].name = host_name
                    self.new_hosts[host_name].site = self.name
                    try:
                        self.new_hosts[host_name].status = self.STATES_MAPPING['hosts'].get(host['attrs']['state'])
                    except KeyError:
                        self.new_hosts[host_name].status = 'UNKNOWN'
                    if int(host['attrs']['state_type']) > 0:  # if state is not SOFT, icinga does not report attempts properly
                        self.new_hosts[host_name].attempt = "{}/{}".format(
                            int(host['attrs']['max_check_attempts']),
                            int(host['attrs']['max_check_attempts']))
                    else:
                        self.new_hosts[host_name].attempt = "{}/{}".format(
                            int(host['attrs']['check_attempt']),
                            int(host['attrs']['max_check_attempts']))
                    self.new_hosts[host_name].last_check = arrow.get(host['attrs']['last_check']).humanize()
                    self.new_hosts[host_name].duration = arrow.get(host['attrs']['previous_state_change']).humanize()
                    self.new_hosts[host_name].status_information = host['attrs']['last_check_result']['output']
                    self.new_hosts[host_name].passiveonly = not(host['attrs']['enable_active_checks'])
                    self.new_hosts[host_name].notifications_disabled = not(host['attrs']['enable_notifications'])
                    self.new_hosts[host_name].flapping = host['attrs']['flapping']
                    self.new_hosts[host_name].acknowledged = bool(host['attrs']['acknowledgement'])
                    self.new_hosts[host_name].scheduled_downtime = bool(host['attrs']['downtime_depth'])
                    self.new_hosts[host_name].status_type = {0: "soft", 1: "hard"}[host['attrs']['state_type']]
                del host_name
            del hosts

        except Exception as e:
            # set checking flag back to False
            self.isChecking = False
            result, error = self.error(sys.exc_info())
            log.exception(e)
            return Result(result=result, error=error)

        # services
        try:
            services = self._get_service_events()
            for service in services:
                new_service = GenericService()
                new_service.host = service['attrs']['host_name']
                new_service.name = service['attrs']['name']
                try:
                    new_service.status = self.STATES_MAPPING['services'].get(service['attrs']['state'])
                except KeyError:
                    new_service.status = 'UNKNOWN'
                if int(service['attrs']['state_type']) > 0:  # if state is not SOFT, icinga does not report attempts properly
                    new_service.attempt = "{}/{}".format(
                        int(service['attrs']['max_check_attempts']),
                        int(service['attrs']['max_check_attempts']))
                else:
                    new_service.attempt = "{}/{}".format(
                        int(service['attrs']['check_attempt']),
                        int(service['attrs']['max_check_attempts']))
                if service['attrs']['last_check_result'] is None:
                    new_service.status_information = 'UNKNOWN'
                else:
                    new_service.status_information = service['attrs']['last_check_result']['output']
                new_service.last_check = arrow.get(service['attrs']['last_check']).humanize()
                new_service.duration = arrow.get(service['attrs']['previous_state_change']).humanize()
                new_service.passiveonly = not(service['attrs']['enable_active_checks'])
                new_service.notifications_disabled = not(service['attrs']['enable_notifications'])
                new_service.flapping = service['attrs']['flapping']
                new_service.acknowledged = bool(service['attrs']['acknowledgement'])
                new_service.scheduled_downtime = bool(service['attrs']['downtime_depth'])
                new_service.status_type = {0: "soft", 1: "hard"}[service['attrs']['state_type']]
                self._insert_service_to_hosts(new_service)
            del services

        except Exception as e:
            log.exception(e)
            # set checking flag back to False
            self.isChecking = False
            result, error = self.error(sys.exc_info())
            return Result(result=result, error=error)

        # dummy return in case all is OK
        return Result()

    def _list_objects(self, object_type, filter):
        """List objects"""
        result = self.fetch_url(
            f'{self.url}/objects/{object_type}?{urllib.parse.urlencode({"filter": filter})}',
            giveback='raw'
        )
        # purify JSON result of unnecessary control sequence \n
        jsonraw, error, status_code = copy.deepcopy(result.result.replace('\n', '')),\
                                      copy.deepcopy(result.error),\
                                      result.status_code

        # check if any error occured
        errors_occured = self.check_for_error(jsonraw, error, status_code)
        # if there are errors return them
        if errors_occured is not None:
            return errors_occured

        jsondict = json.loads(jsonraw)
        return jsondict.get('results', [])

    def _get_service_events(self):
        """
        Suck faulty service events from API
        """
        return self._list_objects('services', 'service.state!=ServiceOK')

    def _get_host_events(self):
        """
        Suck faulty hosts from API
        """
        return self._list_objects('hosts', 'host.state!=0')


    def _trigger_action(self, action, **data):
        """Trigger on action using Icinga2 API"""
        action_data = {k: v for k, v in data.items() if v is not None}
        self.debug(server=self.get_name(), debug=f"Trigger action {action} with data={action_data}")
        try:
            response = self.session.post(
                f'{self.url}/actions/{action}',
                headers={'Accept': 'application/json'},
                json=action_data,
            )
            self.debug(
                server=self.get_name(),
                debug=f"API return on triggering action {action} (status={response.status_code}): "
                f"{response.text}"
            )
            if 200 <= response.status_code <= 299:
                return True
            self.error(f"Fail to trigger action {action}: {response.json().get('status', 'Unknown error')}")
        except IOError as err:
            log.exception("Fail to trigger action %s with data %s", action, data)
            self.error(f"Fail to trigger action {action}: {err}")

    def _set_recheck(self, host, service):
        """
        Please check again Icinga!
        """
        self._trigger_action(
            "reschedule-check",
            type="Service" if service else "Host",
            filter=(
                'host.name == host_name && service.name == service_name'
                if service else 'host.name == host_name'
            ),
            filter_vars=(
                {'host_name': host, 'service_name': service}
                if service else {'host_name': host}
            ),
        )

    # Overwrite function from generic server to add expire_time value
    def set_acknowledge(self, info_dict):
        '''
            different monitors might have different implementations of _set_acknowledge
        '''
        if info_dict['acknowledge_all_services'] is True:
            all_services = info_dict['all_services']
        else:
            all_services = []

        # Make sure expire_time is set
        #if not info_dict['expire_time']:
        #    info_dict['expire_time'] = None

        self._set_acknowledge(info_dict['host'],
                              info_dict['service'],
                              info_dict['author'],
                              info_dict['comment'],
                              info_dict['sticky'],
                              info_dict['notify'],
                              info_dict['persistent'],
                              all_services,
                              info_dict['expire_time'])

    def _set_acknowledge(self, host, service, author, comment, sticky,
                         notify, persistent, all_services=None, expire_time=None):
        '''
        Send acknowledge to monitor server
        '''
        self._trigger_action(
            "acknowledge-problem",
            type="Service" if service else "Host",
            filter=(
                'host.name == host_name && service.name == service_name'
                if service else 'host.name == host_name'
            ),
            filter_vars=(
                {'host_name': host, 'service_name': service}
                if service else {'host_name': host}
            ),
            author=author,
            comment=comment,
            sticky=sticky,
            notify=notify,
            expiry=(
                dateutil.parser.parse(expire_time).timestamp()
                if expire_time else None
            ),
            persistent=persistent,
        )

        if len(all_services) > 0:
            for s in all_services:
                # cheap, recursive solution...
                self._set_acknowledge(host, s, author, comment, sticky, notify, persistent, [], expire_time)

    def _set_submit_check_result(self, host, service, state, comment,
                                 check_output, performance_data):
        '''
        Submit check results
        '''
        self._trigger_action(
            "process-check-result",
            type="Service" if service else "Host",
            filter=(
                'host.name == host_name && service.name == service_name'
                if service else 'host.name == host_name'
            ),
            filter_vars=(
                {'host_name': host, 'service_name': service}
                if service else {'host_name': host}
            ),
            exit_status=self.STATES_MAPPING_REV['services' if service else 'hosts'][state.upper()],
            plugin_output=check_output,
            performance_data=performance_data,
        )

    def _set_downtime(self, host, service, author, comment, fixed, start_time,
                      end_time, hours, minutes):
        """
        Submit downtime
        """
        self._trigger_action(
            "schedule-downtime",
            type="Service" if service else "Host",
            filter=(
                'host.name == host_name && service.name == service_name'
                if service else 'host.name == host_name'
            ),
            filter_vars=(
                {'host_name': host, 'service_name': service}
                if service else {'host_name': host}
            ),
            author=author,
            comment=comment,
            start_time=(
                datetime.datetime.now().timestamp()
                if start_time == '' or start_time == 'n/a'
                else dateutil.parser.parse(start_time).timestamp()
            ),
            end_time=(
                (datetime.datetime.now() + datetime.timedelta(hours=hours, minutes=minutes)).timestamp()
                if end_time == '' or end_time == 'n/a'
                else dateutil.parser.parse(end_time).timestamp()
            ),
            fixed=fixed,
            duration=(
                (hours * 3600 + minutes * 60)
                if not fixed else None
            ),
        )



# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

# Global variables used in different modules

from dataclasses import dataclass

from Nagstamon.config import (conf,
                              OS,
                              OS_MACOS,
                              RESOURCES)
from Nagstamon.helpers import FilesDict
from Nagstamon.qui.dbus import DBus
from Nagstamon.qui.qt import QFont
from Nagstamon.qui.widgets.app import app


# save default font to be able to reset to it
font_default = app.font()

# take global font from conf if it exists
if conf.font != '':
    font = QFont()
    font.fromString(conf.font)
else:
    font = font_default

# always stay in normal weight without any italic
font_icons = QFont('Nagstamon', font.pointSize() + 2, QFont.Weight.Normal, False)

# DBus initialization
dbus_connection = DBus()

# check ECP authentication support availability
try:
    from requests_ecp import HTTPECPAuth
    ecp_available = True
except ImportError:
    ecp_available = False

# flag to keep track of Kerberos availability
kerberos_available = False
if OS == OS_MACOS:
    # requests_gssapi is newer but not available everywhere
    try:
        # extra imports needed to get it compiled on macOS
        import numbers
        import gssapi.raw.cython_converters
        from requests_gssapi import HTTPSPNEGOAuth as HTTPSKerberos
        kerberos_available = True
    except ImportError as error:
        print(error)
else:
    # requests_gssapi is newer but not available everywhere
    try:
        # requests_gssapi needs installation of KfW - Kerberos for Windows
        # requests_kerberoes doesn't
        from requests_kerberos import HTTPKerberosAuth as HTTPSKerberos
        kerberos_available = True
    except ImportError as error:
        print(error)


@dataclass
class StatusWindowProperties:
    """
    storing statuswindow related variables globally available for several classes
    """
    icon_x: int = 0
    icon_y: int = 0
    is_shown: bool = False
    is_shown_timestamp: float = 0.0
    is_hiding_timestamp: float = 0.0
    moving: bool = False
    relative_x: int = 0
    relative_y: int = 0
    status_ok: bool = True
    top: bool = False
    # flag about current notification state
    is_notifying: bool = False
    # current worst state worth a notification
    worst_notification_status: str = 'UP'
    notifying_server: str = ''
    debug_loop_looping = False

# shared status window properties
statuswindow_properties = StatusWindowProperties()

# access to clipboard
clipboard = app.clipboard()

# QBrushes made of QColors for treeview model data() method
# 2 flavours for alternating backgrounds
# filled by create_brushes()
qbrushes = {0: {}, 1: {}}

# store default sounds as buffers to avoid https://github.com/HenriWahl/Nagstamon/issues/578
# meanwhile used as backup copy in case they had been deleted by macOS
# https://github.com/HenriWahl/Nagstamon/issues/578
resource_files = FilesDict(RESOURCES)


# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from io import StringIO
from pathlib import Path

from Nagstamon.config import (conf,
                              RESOURCES,
                              OS,
                              OS_MACOS)

from Nagstamon.qui.constants import ICON
from Nagstamon.qui.globals import statuswindow_properties
from Nagstamon.qui.helpers import hide_macos_dock_icon
from Nagstamon.qui.qt import (QBrush,
                              QListWidgetItem,
                              QObject,
                              QSignalMapper,
                              QSizePolicy,
                              Qt,
                              Signal,
                              Slot,
                              uic,
                              UI_FILE_QT6_QT5_DOWNGRADES)

# make icon status in macOS dock accessible via NSApp, used by set_macos_dock_icon_visible()
if OS == OS_MACOS:
    from AppKit import NSApp


class Dialog(QObject):
    """
    one single dialog
    """
    # send a signal, e.g., to the status window if a dialog pops up
    show_dialog = Signal()

    # signals for macOS dock icon fix
    check_macos_dock_icon_fix_show = Signal()
    check_macos_dock_icon_fix_hide = Signal()

    # dummy toggle dependencies
    TOGGLE_DEPS = {}
    # auxiliary list of checkboxes which HIDE some other widgets if triggered - for example proxy OS settings
    TOGGLE_DEPS_INVERTED = []
    # widgets that might be enabled/disabled depending on a monitor server type
    VOLATILE_WIDGETS = {}
    # names of widgets and their defaults
    WIDGET_NAMES = {}
    # style stuff used by settings dialog for servers/actions listwidget
    GRAY = QBrush(Qt.GlobalColor.gray)

    def __init__(self, dialog):
        QObject.__init__(self)

        # load UI file for manipulation
        ui_file_path = Path(f'{RESOURCES}/qui/{dialog}.ui')
        ui_file_content = ui_file_path.read_text(encoding='utf-8')
        # replace all problematic Qt6 names by Qt5 names to be able to use one single UI file for both Qt5 and Qt6
        for qt6_name, qt5_name in UI_FILE_QT6_QT5_DOWNGRADES.items():
            ui_file_content = ui_file_content.replace(qt6_name, qt5_name)
        # convert string to file-like object for loadUi()
        ui_file_content_io = StringIO(ui_file_content)
        # load UI file from resource
        self.window = uic.loadUi(ui_file_content_io)

        # explicitly set window flags to avoid '?' button on Windows
        self.window.setWindowFlags(Qt.WindowType.WindowCloseButtonHint)

        # hoping to avoid overly large dialogs
        self.window.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)

        # set small titlebar icon
        self.window.setWindowIcon(ICON)

        # treat dialog content after pressing OK button
        if 'button_box' in dir(self.window):
            self.window.button_box.accepted.connect(self.ok)
            self.window.button_box.rejected.connect(self.cancel)

        # QSignalMapper needed to connect all toggle-needing-checkboxes/radiobuttons to one .toggle()-method which
        # decides which sender to use as key in self.TOGGLE_DEPS
        self.signalmapper_toggles = QSignalMapper()

        # try to get and keep focus
        self.window.setWindowModality(Qt.WindowModality.ApplicationModal)

    def initialize(self):
        """
        dummy initialize method
        """
        pass

    @Slot()
    def show(self, tab=0):
        """
        simple show method, to be enriched
        """
        # if running on macOS with disabled dock icon the dock icon might have to be made visible
        # to make Nagstamon accept keyboard input
        self.check_macos_dock_icon_fix_show.emit()

        # in case dock icon is configured invisible in macOS it has to be shown while dialog is shown
        # to be able to get keyboard focus
        if OS == OS_MACOS and \
                conf.icon_in_systray and \
                conf.hide_macos_dock_icon:
            hide_macos_dock_icon(False)

        # tell the world that dialog pops up
        self.show_dialog.emit()

        # reset the window if it only needs smaller screen estate
        self.window.adjustSize()
        self.window.show()
        # make sure the dialog window will be the topmost
        self.window.raise_()
        # hidden dock icon on macOS needs extra activation
        if OS == OS_MACOS and \
                conf.icon_in_systray and \
                conf.hide_macos_dock_icon:
            NSApp.activateIgnoringOtherApps_(True)

    def toggle_visibility(self, checkbox, widgets=None):
        """
        state of checkbox toggles visibility of widgets
        some checkboxes might trigger an inverted behaviour - thus the 'inverted' value
        """
        if widgets is None:
            widgets = []
        if checkbox in self.TOGGLE_DEPS_INVERTED:
            if checkbox.isChecked():
                for widget in widgets:
                    widget.hide()
            else:
                for widget in widgets:
                    widget.show()
        # normal case - click on checkbox activates more options
        else:
            if checkbox.isChecked():
                for widget in widgets:
                    widget.show()
            else:
                for widget in widgets:
                    widget.hide()

    @Slot(str)
    def toggle(self, checkbox):
        """
        change state of the dependant widgets, slot for signals from checkboxes in UI
        """
        # Due to older Qt5 in Ubuntu 14.04 signal mapper has to use strings
        self.toggle_visibility(self.window.__dict__[checkbox],
                               self.TOGGLE_DEPS[self.window.__dict__[checkbox]])

        # adjust dialog window size after UI changes
        self.window.adjustSize()

    def toggle_toggles(self):
        # apply toggle-dependencies between checkboxes as certain widgets
        for checkbox, widgets in self.TOGGLE_DEPS.items():
            # toggle visibility
            self.toggle_visibility(checkbox, widgets)
            # multiplex slot .toggle() by signal-mapping
            # Due to older Qt5 in Ubuntu 14.04 signal mapper has to use strings
            self.signalmapper_toggles.setMapping(checkbox, checkbox.objectName())
            checkbox.toggled.connect(self.signalmapper_toggles.map)
            checkbox.toggled.connect(self.window.adjustSize)

        # finally, map signals with .sender() - [QWidget] is important!
        self.signalmapper_toggles.mappedString[str].connect(self.toggle)

    def fill_list(self, list_widget, config):
        """
        fill list widget with items from config
        """
        for config_item in sorted(config, key=str.lower):
            list_item = QListWidgetItem(config_item)
            if config[config_item].enabled is False:
                list_item.setForeground(self.GRAY)
            list_widget.addItem(list_item)

    @Slot()
    def ok(self):
        """
        as default closes dialog - might be refined, for example, by settings dialog
        """
        self.window.close()
        # en reverse the dock icon might be hidden again after a potential keyboard input
        self.check_macos_dock_icon_fix_show.emit()

    @Slot()
    def cancel(self):
        """
        as default closes dialog - might be refined, for example by settings dialog
        """
        self.window.close()
        # en reverse the dock icon might be hidden again after a potential keyboard input
        self.check_macos_dock_icon_fix_hide.emit()




# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from os import sep

from collections import OrderedDict

from Nagstamon.config import (conf,
                              RESOURCES)
from Nagstamon.qui.constants import (COLORS,
                                     COLOR_STATE_NAMES)
from Nagstamon.qui.globals import statuswindow_properties, font
from Nagstamon.qui.qt import (QSizePolicy,
                              QTimer,
                              QWidget,
                              Signal,
                              Slot)
from Nagstamon.qui.widgets.draggables import DraggableLabel
from Nagstamon.qui.widgets.layout import HBoxLayout
from Nagstamon.qui.widgets.nagstamon_logo import NagstamonLogo
from Nagstamon.servers import servers, get_errors


class StatusBarLabel(DraggableLabel):
    """
    one piece of the status bar labels for one state
    """

    # yell if statusbar is moved
    window_moved = Signal()

    # needed for popup after hover
    mouse_entered = Signal()

    # needed for popup after click
    mouse_pressed = Signal()
    mouse_released = Signal()
    # needed to close window in some configurations
    mouse_released_in_window = Signal()

    def __init__(self, state, parent=None):
        DraggableLabel.__init__(self, parent=parent)
        self.setStyleSheet(f'''padding-left: 1px;
                               padding-right: 1px;
                               color: {conf.__dict__[f'color_{state.lower()}_text']};
                               background-color: {conf.__dict__[f'color_{state.lower()}_background']};
                            ''')
        # just let labels grow as much as they need
        self.setSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Maximum)

        # hidden per default
        self.hide()

        # default text - only useful in case of OK Label
        self.setText(state)

        # number of hosts/services of this state
        self.number = 0

        # store state of label to access long state names in .summarize_states()
        self.state = state

    @Slot()
    def invert(self):
        self.setStyleSheet(f'''padding-left: 1px;
                               padding-right: 1px;
                               color: {conf.__dict__[f'color_{self.state.lower()}_background']};
                               background-color: {conf.__dict__[f'color_{self.state.lower()}_text']};
                            ''')

    @Slot()
    def reset(self):
        self.setStyleSheet(f'''padding-left: 1px;
                               padding-right: 1px;
                               color: {conf.__dict__[f'color_{self.state.lower()}_text']};
                               background-color: {conf.__dict__[f'color_{self.state.lower()}_background']};
                            ''')


class StatusBar(QWidget):
    """
    status bar for short display of problems
    """

    # send signal to statuswindow
    resize = Signal()

    # needed to maintain flashing labels
    labels_invert = Signal()
    labels_reset = Signal()

    def __init__(self, parent=None):
        QWidget.__init__(self, parent=parent)

        self.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)

        self.hbox = HBoxLayout(spacing=0, parent=parent)
        self.setLayout(self.hbox)

        # define labels first to get their size for svg logo dimensions
        self.color_labels = OrderedDict()
        self.color_labels['OK'] = StatusBarLabel('OK', parent=parent)

        for state in COLORS:
            self.color_labels[state] = StatusBarLabel(state, parent=parent)
            self.labels_invert.connect(self.color_labels[state].invert)
            self.labels_reset.connect(self.color_labels[state].reset)

        # label for error message(s)
        self.label_message = StatusBarLabel('error', parent=parent)
        self.labels_invert.connect(self.label_message.invert)
        self.labels_reset.connect(self.label_message.reset)

        # derive logo dimensions from status label
        self.logo = NagstamonLogo(f'{RESOURCES}{sep}nagstamon_logo_bar.svg',
                                  self.color_labels['OK'].fontMetrics().height(),
                                  self.color_labels['OK'].fontMetrics().height(),
                                  parent=parent)

        # add logo
        self.hbox.addWidget(self.logo)

        # label for error messages
        self.hbox.addWidget(self.label_message)
        self.label_message.hide()

        # add state labels
        self.hbox.addWidget(self.color_labels['OK'])
        for state in COLORS:
            self.hbox.addWidget(self.color_labels[state])

        # timer for singleshots for flashing
        self.timer = QTimer()

        self.adjust_size()

    @Slot()
    def summarize_states(self):
        """
        display summaries of states in statusbar
        """
        # initial zeros
        for label in self.color_labels.values():
            label.number = 0

        # only count numbers of enabled monitor servers
        for server in (filter(lambda s: s.enabled, servers.values())):
            for state in COLORS:
                self.color_labels[state].number += server.__dict__[state.lower()]

        # summarize all numbers - if all_numbers keeps 0 everything seems to be OK
        all_numbers = 0

        # repaint colored labels or hide them if necessary
        for label in self.color_labels.values():
            if label.number == 0:
                label.hide()
            else:
                label.setText(' '.join((str(label.number),
                                        COLOR_STATE_NAMES[label.state][conf.long_display])))
                label.show()
                label.adjustSize()
                all_numbers += label.number

        if all_numbers == 0 and not get_errors() and not self.label_message.isVisible():
            self.color_labels['OK'].show()
            self.color_labels['OK'].adjustSize()
        else:
            self.color_labels['OK'].hide()

        # fix size after refresh - better done here to avoid ugly artefacts
        hint = self.sizeHint()
        self.setMaximumSize(hint)
        self.setMinimumSize(hint)
        del hint
        # tell statuswindow its size might be adjusted
        self.resize.emit()

    @Slot()
    def flash(self):
        """
        send color inversion signal to labels
        """
        # only if currently a notification is necessary
        if statuswindow_properties.is_notifying:
            self.labels_invert.emit()
            # fire up  a singleshot to reset color soon
            self.timer.singleShot(500, self.reset)

    @Slot()
    def reset(self):
        """
        tell labels to set original colors
        """
        self.labels_reset.emit()
        # only if currently a notification is necessary
        if statuswindow_properties.is_notifying:
            # even later call itself to invert colors as flash
            self.timer.singleShot(500, self.flash)

    @Slot()
    def adjust_size(self):
        """
        apply new size of widgets, especially Nagstamon logo
        run through all labels to the the max height in case not all labels
        are shown at the same time - which is very likely the case
        """
        # take height for logo
        # height = 0

        # run through labels to set font and get height for logo
        for label in self.color_labels.values():
            label.setFont(font)
        #    if label.fontMetrics().height() > height:
        #        height = label.fontMetrics().height()

        self.label_message.setFont(font)
        height = self.label_message.sizeHint().height()

        # adjust logo size to fit to label size - due to being a square height and width are the same
        self.logo.adjust_size(height, height)

        # avoid flickering/artefact by updating immediately
        self.summarize_states()

    @Slot(str)
    def set_error(self, message):
        """
        display error message if any error exists
        """
        self.label_message.setText(message)
        self.label_message.show()

    @Slot()
    def reset_error(self):
        """
        delete error message if there is no error
        """
        if not get_errors():
            self.label_message.setText('')
            self.label_message.hide()



# encoding: utf-8

# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

# LibreNMS integration for Nagstamon
#
# This Server class connects against LibreNMS API.
# The monitor URL in the setup should be something like
# http://librenms.example.com
#
# Authentication: Use API token as password field, username can be empty
# API token can be generated in LibreNMS under: User > API Settings

import sys
import json
from datetime import datetime

from Nagstamon.config import conf
from Nagstamon.objects import (GenericHost,
                               GenericService,
                               Result)
from Nagstamon.servers.Generic import GenericServer
from Nagstamon.helpers import webbrowser_open


class LibreNMSServer(GenericServer):
    """
    special treatment for LibreNMS API
    """
    TYPE = 'LibreNMS'

    # LibreNMS actions
    MENU_ACTIONS = ['Monitor', 'Acknowledge']
    BROWSER_URLS = {
        'monitor': '$MONITOR$',
        'hosts': '$MONITOR$/devices',
        'services': '$MONITOR$/alerts',
        'history': '$MONITOR$/eventlog'
    }

    # API paths
    API_PATH_ALERTS = "/api/v0/alerts"
    API_PATH_DEVICES = "/api/v0/devices"

    # Status mapping from LibreNMS severity to Nagios-style states
    # LibreNMS severities: ok, warning, critical
    SEVERITY_MAPPING = {
        'ok': 'OK',
        'warning': 'WARNING',
        'critical': 'CRITICAL'
    }

    def __init__(self, **kwds):
        GenericServer.__init__(self, **kwds)

        # Store device information for alert enrichment
        self.devices = {}

        # Cache for services (by device_id)
        self.services_cache = {}

        # Option to treat service status as alerts
        # When enabled, non-OK services will be shown even without alert rules
        self.treat_services_as_alerts = False

    def init_http(self):
        """
        things to do if HTTP is not initialized
        """
        GenericServer.init_http(self)

        # prepare for JSON and set API token header
        # LibreNMS uses X-Auth-Token header for authentication
        if self.session and hasattr(self.session, 'headers'):
            self.session.headers.update({
                'Accept': 'application/json',
                'Content-Type': 'application/json',
                'X-Auth-Token': self.password  # API token goes in X-Auth-Token header
            })

    def init_config(self):
        """
        dummy init_config, called at thread start
        """
        pass

    def _get_device_display_name(self, device_id, hostname_fallback):
        """
        Get the display name for a device from cache or API
        """
        # Check if we already have this device cached
        if device_id in self.devices:
            return self.devices[device_id].get('display', hostname_fallback)

        # Fetch device details from API
        try:
            result = self.fetch_url(
                f"{self.monitor_url}/api/v0/devices/{device_id}",
                giveback="json"
            )

            if result.result and result.result.get('status') == 'ok':
                device_data = result.result.get('devices', [{}])[0]
                # Cache the device info
                self.devices[device_id] = device_data

                # Try different fields in order of preference
                # display > sysName > hostname
                display_name = (device_data.get('display') or
                               device_data.get('sysName') or
                               device_data.get('hostname') or
                               hostname_fallback)

                if conf.debug_mode:
                    self.debug(server=self.get_name(),
                              debug=f"Device {device_id}: display='{display_name}' (from {device_data.get('hostname', 'N/A')})")

                return display_name
        except Exception:
            if conf.debug_mode:
                self.debug(server=self.get_name(),
                          debug=f"Could not fetch device {device_id}, using fallback: {hostname_fallback}")

        return hostname_fallback

    def _get_alert_details(self, alert_id):
        """
        Get detailed alert information including rule name
        """
        try:
            result = self.fetch_url(
                f"{self.monitor_url}/api/v0/alerts/{alert_id}",
                giveback="json"
            )

            if result.result and result.result.get('status') == 'ok':
                alerts = result.result.get('alerts', [])
                if alerts:
                    return alerts[0]
        except Exception:
            if conf.debug_mode:
                self.debug(server=self.get_name(),
                          debug=f"Could not fetch alert details for {alert_id}")

        return None

    def _get_rule_name(self, rule_id):
        """
        Get the rule name from cache or API
        """
        # Use a rules cache dictionary
        if not hasattr(self, 'rules_cache'):
            self.rules_cache = {}

        # Check cache first
        if rule_id in self.rules_cache:
            return self.rules_cache[rule_id]

        # Fetch rule details from API
        try:
            result = self.fetch_url(
                f"{self.monitor_url}/api/v0/rules/{rule_id}",
                giveback="json"
            )

            if result.result and result.result.get('status') == 'ok':
                rules = result.result.get('rules', [])
                if rules:
                    rule_name = rules[0].get('name', f'Rule {rule_id}')
                    # Cache it
                    self.rules_cache[rule_id] = rule_name
                    return rule_name
        except Exception:
            if conf.debug_mode:
                self.debug(server=self.get_name(),
                          debug=f"Could not fetch rule {rule_id}")

        return f"Rule {rule_id}"

    def _get_service_details(self, device_id, service_name):
        """
        Find a matching service for the alert and return its details
        Returns service_message if found
        """
        # Check if we have services cached for this device
        if device_id not in self.services_cache:
            # Fetch services for this device
            try:
                result = self.fetch_url(
                    f"{self.monitor_url}/api/v0/services",
                    giveback="json"
                )

                if result.result and result.result.get('status') == 'ok':
                    services = result.result.get('services', [])

                    # Build a device_id -> services mapping
                    for svc_list in services:
                        for svc in svc_list:
                            dev_id = svc.get('device_id')
                            if dev_id not in self.services_cache:
                                self.services_cache[dev_id] = []
                            self.services_cache[dev_id].append(svc)

                    if conf.debug_mode:
                        self.debug(server=self.get_name(),
                                  debug=f"Cached services for all devices")
            except Exception:
                if conf.debug_mode:
                    self.debug(server=self.get_name(),
                              debug=f"Could not fetch services")
                return None

        # Now find matching service by name
        device_services = self.services_cache.get(device_id, [])
        for svc in device_services:
            if svc.get('service_name', '').lower() in service_name.lower() or \
               service_name.lower() in svc.get('service_name', '').lower():
                return svc.get('service_message', '')

        return None

    def _get_status(self):
        """
        Get status from LibreNMS Server
        """
        # Reset new_hosts dictionary
        self.new_hosts = {}

        try:
            # Get alerts from LibreNMS API
            # state=1 means "alert" status (active alerts)
            # We'll get both active (state=1) and acknowledged (state=2) alerts
            result = self.fetch_url(
                self.monitor_url + self.API_PATH_ALERTS,
                giveback="json"
            )

            data = result.result
            error = result.error
            status_code = result.status_code

            # check if any error occurred
            errors_occurred = self.check_for_error(data, error, status_code)
            if errors_occurred is not None:
                return errors_occurred

            if conf.debug_mode:
                self.debug(server=self.get_name(),
                          debug=f"Fetched {len(data.get('alerts', []))} alerts from LibreNMS")

            # Process each alert
            alerts = data.get('alerts', [])

            for alert in alerts:
                # Skip alerts that are in "ok" state (state=0)
                # state: 0=ok, 1=alert, 2=acknowledged
                alert_state = int(alert.get('state', 0))
                if alert_state == 0:
                    continue

                # Get device information
                device_id = alert.get('device_id')
                hostname_raw = alert.get('hostname', f'device_{device_id}')

                # Get the display name from device API (cached)
                hostname = self._get_device_display_name(device_id, hostname_raw)

                # Get alert details
                alert_id = alert.get('id')
                rule_id = alert.get('rule_id')

                # LibreNMS provides a 'name' field - use it directly!
                service_name = alert.get('name', f'Alert {alert_id}')

                # Get timestamp and calculate duration
                timestamp = alert.get('timestamp', '')

                # Get severity directly from alert (warning, critical, ok)
                severity = alert.get('severity', 'unknown').upper()

                # Map LibreNMS severity to Nagstamon status
                status_mapping = {
                    'OK': 'OK',
                    'WARNING': 'WARNING',
                    'CRITICAL': 'CRITICAL',
                    'UNKNOWN': 'UNKNOWN'
                }
                status = status_mapping.get(severity, 'CRITICAL')

                # Build informative status message
                status_parts = []

                # Try to get detailed service message from services API
                # This only works for service-based alerts (NRPE, custom checks)
                # Metric-based alerts won't have a matching service
                service_message = self._get_service_details(device_id, service_name)

                # If we found a matching service, prioritize its detailed message
                if service_message:
                    status_parts.append(f"Service: {service_message}")
                    if conf.debug_mode:
                        self.debug(server=self.get_name(),
                                  debug=f"Found service match for alert {alert_id}: {service_name}")
                else:
                    # No service match - this is likely a metric-based alert
                    # Use alert fields directly
                    if conf.debug_mode:
                        self.debug(server=self.get_name(),
                                  debug=f"No service match for alert {alert_id} - using metric data")

                # Add info field if available (for metric alerts, this is important!)
                alert_info = alert.get('info', '').strip()
                if alert_info and alert_info not in str(service_message):
                    status_parts.append(alert_info)

                # Add proc field if available
                alert_proc = alert.get('proc', '').strip()
                if alert_proc and alert_proc not in str(service_message):
                    status_parts.append(alert_proc)

                # Add notes if available
                alert_notes = alert.get('notes', '').strip()
                if alert_notes:
                    status_parts.append(f"Notes: {alert_notes}")

                # Add acknowledgment note if exists
                alert_note = alert.get('note', '').strip()
                if alert_note:
                    # Just show the last line of the note (most recent)
                    last_note = alert_note.split('\n')[-1] if '\n' in alert_note else alert_note
                    status_parts.append(f"Ack: {last_note}")

                # If we still don't have any info, show timestamp
                if not status_parts and timestamp:
                    status_parts.append(f"Active since {timestamp}")

                # Final fallback
                if not status_parts:
                    status_parts.append(f"Alert ID: {alert_id}")

                status_information = " | ".join(status_parts)

                # Create service object
                service = GenericService()
                service.host = hostname
                service.name = service_name
                service.server = self.name
                service.status = status
                service.last_check = timestamp
                service.duration = self._calculate_duration(timestamp)
                service.attempt = 'n/a'

                # Check if acknowledged (state=2)
                service.acknowledged = (alert_state == 2)

                # Store alert_id for acknowledge operations
                service.real_name = str(alert_id)

                # Set the status information
                service.status_information = status_information

                # Create host if not exists
                if hostname not in self.new_hosts:
                    self.new_hosts[hostname] = GenericHost()
                    self.new_hosts[hostname].name = hostname
                    self.new_hosts[hostname].server = self.name
                    self.new_hosts[hostname].status = 'UP'

                # Add service to host
                self.new_hosts[hostname].services[service_name] = service

                if conf.debug_mode:
                    self.debug(server=self.get_name(),
                              debug=f"Added alert {alert_id} for {hostname}: {service_name}")

            # If "treat services as alerts" is enabled, also add non-OK services
            if self.treat_services_as_alerts:
                if conf.debug_mode:
                    self.debug(server=self.get_name(),
                              debug="Processing services as alerts...")

                self._add_services_as_alerts()

        except Exception:
            # set checking flag back to False
            self.isChecking = False
            result, error = self.error(sys.exc_info())
            return Result(result=result, error=error)

        # dummy return in case all is OK
        return Result()

    def _add_services_as_alerts(self):
        """
        Add non-OK services as alerts (when treat_services_as_alerts is enabled)
        This is called after alerts are processed to avoid duplicates
        """
        # Track which device/service combinations we already have from alerts
        existing_alerts = set()
        for hostname, host_obj in self.new_hosts.items():
            for service_name in host_obj.services.keys():
                existing_alerts.add((host_obj.name, service_name))

        # Fetch all services if not cached
        if not self.services_cache:
            try:
                result = self.fetch_url(
                    f"{self.monitor_url}/api/v0/services",
                    giveback="json"
                )

                if result.result and result.result.get('status') == 'ok':
                    services = result.result.get('services', [])
                    for svc_list in services:
                        for svc in svc_list:
                            dev_id = svc.get('device_id')
                            if dev_id not in self.services_cache:
                                self.services_cache[dev_id] = []
                            self.services_cache[dev_id].append(svc)
            except Exception:
                if conf.debug_mode:
                    self.debug(server=self.get_name(),
                              debug="Could not fetch services for treat_services_as_alerts")
                return

        # Process non-OK services
        for device_id, services_list in self.services_cache.items():
            for svc in services_list:
                # Skip OK services (status=0)
                service_status = svc.get('service_status', 0)
                if service_status == 0:
                    continue

                # Skip disabled or ignored services
                if svc.get('service_disabled', 0) == 1 or svc.get('service_ignore', 0) == 1:
                    continue

                # Get device display name
                hostname_raw = f"device_{device_id}"
                hostname = self._get_device_display_name(device_id, hostname_raw)

                service_name_raw = svc.get('service_name', 'Unknown Service')
                # Add marker to distinguish services from alerts
                service_name = f"[S] {service_name_raw}"

                # Check if we already have this from an alert (deduplicate)
                # Use raw name for deduplication check
                if (hostname, service_name_raw) in existing_alerts or (hostname, service_name) in existing_alerts:
                    if conf.debug_mode:
                        self.debug(server=self.get_name(),
                                  debug=f"Skipping duplicate: {hostname}/{service_name}")
                    continue

                # Map service status to Nagstamon status
                status_map = {
                    0: 'OK',
                    1: 'WARNING',
                    2: 'CRITICAL',
                    3: 'UNKNOWN'
                }
                status = status_map.get(service_status, 'UNKNOWN')

                # Create service object
                service = GenericService()
                service.host = hostname
                service.name = service_name
                service.server = self.name
                service.status = status
                service.last_check = 'n/a'
                service.duration = 'n/a'
                service.attempt = 'n/a'
                service.acknowledged = False

                # Use service message as status information
                service.status_information = svc.get('service_message', 'No message')

                # Store service_id for potential future operations
                service.real_name = f"service_{svc.get('service_id')}"

                # Create host if not exists
                if hostname not in self.new_hosts:
                    self.new_hosts[hostname] = GenericHost()
                    self.new_hosts[hostname].name = hostname
                    self.new_hosts[hostname].server = self.name
                    self.new_hosts[hostname].status = 'UP'

                # Add service to host
                self.new_hosts[hostname].services[service_name] = service

                if conf.debug_mode:
                    self.debug(server=self.get_name(),
                              debug=f"Added service as alert: {hostname}/{service_name} ({status})")

    def _calculate_duration(self, timestamp):
        """
        Calculate duration from timestamp to now
        timestamp format from LibreNMS: "2014-12-11 14:40:02"
        """
        try:
            if not timestamp:
                return "n/a"

            # Parse timestamp
            alert_time = datetime.strptime(timestamp, "%Y-%m-%d %H:%M:%S")
            now = datetime.now()
            duration = now - alert_time

            days = duration.days
            hours = duration.seconds // 3600
            minutes = (duration.seconds % 3600) // 60
            seconds = duration.seconds % 60

            if days > 0:
                return f"{days}d {hours}h {minutes:02d}m"
            elif hours > 0:
                return f"{hours}h {minutes:02d}m {seconds:02d}s"
            elif minutes > 0:
                return f"{minutes}m {seconds:02d}s"
            else:
                return f"{seconds}s"
        except Exception:
            return "n/a"

    def _set_acknowledge(self, host, service, author, comment, sticky, notify, persistent, all_services=None):
        """
        Acknowledge an alert in LibreNMS
        """
        try:
            # Find the alert ID from the service
            if host in self.hosts and service in self.hosts[host].services:
                service_obj = self.hosts[host].services[service]
                alert_id = service_obj.real_name

                # LibreNMS acknowledge endpoint: PUT /api/v0/alerts/:id
                url = f"{self.monitor_url}/api/v0/alerts/{alert_id}"

                # Prepare acknowledge data
                # note and until_clear are optional parameters
                data = {}
                if comment:
                    data['note'] = comment
                # until_clear: if false, alert will re-alert if it changes
                # We'll set to false by default for Nagstamon behavior
                data['until_clear'] = False

                # Make PUT request using session directly
                # since fetch_url doesn't support PUT method
                # The session already has verify setting from init_http()
                if self.session:
                    response = self.session.put(
                        url,
                        json=data,
                        timeout=self.timeout,
                        verify=self.session.verify  # Explicitly use session's verify setting
                    )

                    if conf.debug_mode:
                        self.debug(server=self.get_name(),
                                  debug=f"Acknowledged alert {alert_id}: status {response.status_code}")

                    # Check for errors
                    if response.status_code >= 400:
                        error_msg = f"Failed to acknowledge alert {alert_id}: HTTP {response.status_code}"
                        self.error(error_msg)
                        return Result(result=False, error=error_msg)

        except Exception:
            result, error = self.error(sys.exc_info())
            return Result(result=result, error=error)

        return Result()

    def open_monitor(self, host, service=''):
        """
        open monitor for specific alert
        """
        if service and host in self.hosts and service in self.hosts[host].services:
            service_obj = self.hosts[host].services[service]
            alert_id = service_obj.real_name
            url = f"{self.monitor_url}/alerts/{alert_id}"
        else:
            # Open general alerts page
            url = f"{self.monitor_url}/alerts"

        if conf.debug_mode:
            self.debug(server=self.get_name(), host=host, service=service,
                      debug=f"Opening monitor web page {url}")

        webbrowser_open(url)

    def open_monitor_webpage(self):
        """
        open monitor from systray/toparea context menu
        """
        if conf.debug_mode:
            self.debug(server=self.get_name(),
                      debug=f"Opening monitor web page {self.monitor_url}")
        webbrowser_open(self.monitor_url)



#!/usr/bin/python
# adapted from the zabbix.py code

# Copyright (C) 2016 Jake Murphy <jake.murphy@faredge.com.au> Far Edge Technology
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

import sys
import urllib
import base64
import time
import copy
import datetime
import traceback
from datetime import datetime

from Nagstamon.config import conf
from Nagstamon.objects import (GenericHost, GenericService, Result)
from Nagstamon.servers.Generic import GenericServer
from Nagstamon.thirdparty.zenoss_api import ZenossAPI

class ZenossServer(GenericServer):
    
    TYPE = 'Zenoss'
    zapi = None

    SEVERITY_MAP = {0: 'OK',
                1: 'UNKNOWN',
                2: 'UNKNOWN',
                3: 'WARNING',
                4: 'WARNING',
                5: 'CRITICAL'}

    MENU_ACTIONS = ['Monitor', 'Acknowledge']

    def __init__(self, **kwds):
        GenericServer.__init__(self, **kwds)

        # Prepare all urls needed by nagstamon 
        self.urls = {}
        self.statemap = {}

        self.server = Server()
        if ":" in conf.servers[self.get_name()].monitor_url:
            self.server.server_url, self.server.server_port = conf.servers[self.get_name()].monitor_url.split(':')
        else:
            self.server.server_url = conf.servers[self.get_name()].monitor_url
            self.server.server_port = 8080 #the default is 8080

        self.server.username = conf.servers[self.get_name()].username
        self.server.password = conf.servers[self.get_name()].password
        
        # Entries for monitor default actions in context menu
        self.MENU_ACTIONS = ["Acknowledge"]
    
    def _zlogin(self):
        try:
            self.zapi = ZenossAPI(Server=self.server)
        except Exception:
            result, error = self.error(sys.exc_info())
            return Result(result=result, error=error)
    
    def _get_status(self):
        nagitems = {"services":[], "hosts":[]}

        self.new_hosts = dict()

        try:
            hosts = self._get_all_events()
            
            if 'events' in hosts:
                hosts = hosts['events']

                for host in hosts:
                    n = dict()
                    n['evid'] = host['evid']
                    n['host'] = host['device']['text']
                    n['service'] = host['eventClass']['text']

                    n['status'] = self.SEVERITY_MAP.get(host['severity'])
                    n['last_check'] = host['lastTime']
                
                    duration = self._calc_duration(host['firstTime'], host['lastTime'])
                    if (duration == None):
                        continue #Zenoss needs a length to cause an error
                    n['duration'] = duration

                    n["status_information"] = host['message']
                    n["attempt"] = str(host['count'])+"/1" # needs a / with a number on either side to work

                    n["passiveonly"] = False
                    n["notifications_disabled"] = False
                    n["flapping"] = False
                    n["acknowledged"] = (host['eventState'] == 'Acknowledged')
                    n["scheduled_downtime"] = False
              
                    nagitems["hosts"].append(n)

                    new_host = n["host"]
                    if not new_host in self.new_hosts:
                        self.new_hosts[new_host] = GenericHost()
                        self.new_hosts[new_host].name = new_host
                    
                    if not new_host in self.new_hosts[new_host].services:
                        
                        new_service = new_host
                        self.new_hosts[new_host].services[new_service] = GenericService()
                        
                        self.new_hosts[new_host].services[new_service].host = new_host
                        self.new_hosts[new_host].services[new_service].evid = n['evid']
                        self.new_hosts[new_host].services[new_service].name = n["service"]
                    
                        self.new_hosts[new_host].services[new_service].server = self.name
                        self.new_hosts[new_host].services[new_service].status = n["status"]
                        self.new_hosts[new_host].services[new_service].last_check = n["last_check"]
                    
                        self.new_hosts[new_host].services[new_service].duration = n["duration"]
                        self.new_hosts[new_host].services[new_service].status_information= n["status_information"].encode("utf-8")
                        self.new_hosts[new_host].services[new_service].attempt = n["attempt"]

                        self.new_hosts[new_host].services[new_service].passiveonly = n["passiveonly"]
                        self.new_hosts[new_host].services[new_service].notifications_disabled = n["notifications_disabled"]
                        self.new_hosts[new_host].services[new_service].flapping = n["flapping"]
                        self.new_hosts[new_host].services[new_service].acknowledged = n["acknowledged"]
                        self.new_hosts[new_host].services[new_service].scheduled_downtime = n["scheduled_downtime"]
                    del n
                
        except:
            self.isChecking = False
            result, error = self.error(sys.exc_info())
            print(traceback.format_exc())
            return Result(result=result, error=error)

        
        del nagitems
        return Result(error="")
    
    def get_username(self):
        return str(self.server.username)
    def get_password(self):
        return str(self.server.password)
    
    def set_acknowledge(self, info_dict):
        if info_dict['host'] in self.hosts:
            evid = self.hosts[info_dict['host']].services[info_dict['host']].evid
            self.zapi.set_event_ack(evid)

    def _open_browser(self, url):
        webbrowser.open(self.monitor_url)

    def _get_all_events(self):
        if self.zapi is None:
            self._zlogin()

        events = self.zapi.get_event()
        return events

    ##http://stackoverflow.com/questions/538666/python-format-timedelta-to-string
    def _calc_duration(self, startStr, endStr): # like: '2016-10-2213: 53: 43' (that day/hour gap..)
        start = datetime.strptime(startStr, '%Y-%m-%d %H:%M:%S')
        end = datetime.strptime(endStr, '%Y-%m-%d %H:%M:%S')
        
        sec = (int)((end - start).total_seconds())
        
        days, rem = divmod(sec, 60*60*24)
        hours, rem = divmod(rem, 60*60)
        mins, sec = divmod(rem, 60)
        if (days == 0 and hours == 0 and mins == 0 and sec == 0):
            return None
        return '%sd %sh %sm %ss' % (days,hours,mins,sec)
    
    #Note these methods are invalid for the zenoss api that this uses
    def set_recheck(self, info_dict):
        pass

    def set_downtime(self, host, service, author, comment, fixed, start_time, end_time, hours, minutes):
        pass

    def set_submit_check_result(self, info_dict):
        pass
       
    def get_start_end(self, host):
        pass


class Server:
    #server object for api configuration connections

    def __init__(self):
        self.server_url = ""
        self.server_port = ""
        self.username = ""
        self.password = ""



# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from Nagstamon.qui.helpers import check_servers
from Nagstamon.qui.qt import Slot
from Nagstamon.qui.dialogs.dialog import Dialog


class DialogServerMissing(Dialog):
    """
    small dialog to ask about disabled ot not configured servers
    """

    def __init__(self):
        Dialog.__init__(self, 'dialog_server_missing')

        # hide dialog when server is to be created or enabled
        self.window.button_create_server.clicked.connect(self.window.hide)
        self.window.button_enable_server.clicked.connect(self.window.hide)
        self.window.button_ignore.clicked.connect(self.ok)
        # simply hide the window if ignore button chosen
        self.window.button_ignore.clicked.connect(self.window.hide)
        self.window.button_ignore.clicked.connect(self.cancel)
        # bye bye if exit button was pressed - exit is connected in qui/__init__.py
        self.window.button_exit.clicked.connect(self.window.hide)
        # check if servers are configured or enabled and show dialog if not
        check_servers.checked.connect(self.show)
        check_servers.checked.connect(self.initialize)

    @Slot(str)
    def initialize(self, mode='no_server'):
        """
        use dialog for missing and not enabled servers, depending on mode
        """
        if mode == 'no_server':
            self.window.label_no_server_configured.show()
            self.window.label_no_server_enabled.hide()
            self.window.button_enable_server.hide()
            self.window.button_create_server.show()
        else:
            self.window.label_no_server_configured.hide()
            self.window.label_no_server_enabled.show()
            self.window.button_enable_server.show()
            self.window.button_create_server.hide()


# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
from base64 import b64encode
from os import sep

from Nagstamon.config import AppInfo, RESOURCES
from Nagstamon.qui.constants import SPACE
from Nagstamon.qui.globals import app
from Nagstamon.qui.qt import (QAction,
                              QByteArray,
                              QHBoxLayout,
                              QIcon,
                              QPainter,
                              QPixmap,
                              QPalette,
                              QSizePolicy,
                              QSvgRenderer,
                              Qt,
                              QXmlStreamReader,
                              QWidget,
                              Signal,
                              Slot)
from Nagstamon.qui.widgets.buttons import (Button,
                                           CSS_CLOSE_BUTTON,
                                           PushButtonHamburger)
from Nagstamon.qui.widgets.combobox_servers import (ComboBoxServers)
from Nagstamon.qui.widgets.draggables import DraggableLabel
from Nagstamon.qui.widgets.layout import HBoxLayout
from Nagstamon.qui.widgets.menu import MenuAtCursor
from Nagstamon.qui.widgets.nagstamon_logo import NagstamonLogo


class TopArea(QWidget):
    """
    top area of status window
    """

    mouse_entered = Signal()

    def __init__(self, parent=None):
        QWidget.__init__(self)

        self.hbox = HBoxLayout(spacing=SPACE, parent=self)  # top HBox containing buttons
        self.hbox.setSizeConstraint(QHBoxLayout.SizeConstraint.SetMinimumSize)

        self.icons = dict()
        self.create_icons()

        # top button box
        self.logo = NagstamonLogo(self.icons['nagstamon_logo_toparea'], width=150, height=42, parent=self)
        self.label_version = DraggableLabel(text=AppInfo.VERSION, parent=self)
        self.label_empty_space = DraggableLabel(text='', parent=self)
        self.label_empty_space.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Ignored)
        self.combobox_servers = ComboBoxServers(parent=self)
        self.button_filters = Button("Filters", parent=self)
        self.button_recheck_all = Button("Recheck all", parent=self)
        self.button_refresh = Button("Refresh", parent=self)
        self.button_settings = Button("Settings", parent=self)

        # fill default order fields combobox with server names
        self.combobox_servers.fill()

        # hamburger menu
        self.button_hamburger_menu = PushButtonHamburger()
        self.button_hamburger_menu.setIcon(self.icons['menu'])
        self.hamburger_menu = MenuAtCursor()
        self.action_exit = QAction("Exit", self)
        self.hamburger_menu.addAction(self.action_exit)
        self.button_hamburger_menu.setMenu(self.hamburger_menu)

        # X
        self.button_close = Button()
        self.button_close.setIcon(self.icons['close'])
        self.button_close.setStyleSheet(CSS_CLOSE_BUTTON)

        self.hbox.addWidget(self.logo)
        self.hbox.addWidget(self.label_version)
        self.hbox.addWidget(self.label_empty_space)
        self.hbox.addWidget(self.combobox_servers)
        self.hbox.addWidget(self.button_filters)
        self.hbox.addWidget(self.button_recheck_all)
        self.hbox.addWidget(self.button_refresh)
        self.hbox.addWidget(self.button_settings)
        self.hbox.addWidget(self.button_hamburger_menu)
        self.hbox.addWidget(self.button_close)

        self.setLayout(self.hbox)

    def enterEvent(self, event):
        # unlock statuswindow if pointer touches statusbar
        self.mouse_entered.emit()

    @Slot()
    def create_icons(self):
        """
        create icons from template, applying colors
        """
        # get rgb values of current foreground color to be used for SVG icons (menu)
        r, g, b, a = app.palette().color(QPalette.ColorRole.Text).getRgb()

        for icon in 'nagstamon_logo_toparea', 'close', 'menu':
            # get template from file
            svg_template_file = open(f'{RESOURCES}{sep}{icon}_template.svg')
            svg_template_xml = svg_template_file.readlines()

            # current SVG XML for state icon, derived from svg_template_cml
            svg_icon_xml = list()

            # replace dummy text and background colors with configured ones
            for line in svg_template_xml:
                line = line.replace('fill:#ff00ff', 'fill:#{0:x}{1:x}{2:x}'.format(r, g, b))
                svg_icon_xml.append(line)

            # create XML stream of SVG
            svg_xml_stream = QXmlStreamReader(''.join(svg_icon_xml))

            # create renderer for SVG and put SVG XML into renderer
            svg_renderer = QSvgRenderer(svg_xml_stream)
            # pixmap to be painted on - arbitrarily choosen 128x128 px
            svg_pixmap = QPixmap(128, 128)
            # fill transparent backgound
            svg_pixmap.fill(Qt.GlobalColor.transparent)
            # initiate painter which paints onto paintdevice pixmap
            svg_painter = QPainter(svg_pixmap)
            # render svg to pixmap
            svg_renderer.render(svg_painter)
            # close painting
            svg_painter.end()

            # two ways...
            if icon == 'nagstamon_logo_toparea':
                # first get a base64 version of the SVG
                svg_base64 = b64encode(bytes(''.join(svg_icon_xml), 'utf8'))
                # create a QByteArray for NagstamonLogo aka QSvgWidget
                svg_bytes = QByteArray.fromBase64(svg_base64)
                self.icons[icon] = svg_bytes
            else:
                # put pixmap into icon
                self.icons[icon] = QIcon(svg_pixmap)
